"""SDGym module to handle datasets."""

import logging
import os
from pathlib import Path

import appdirs
import numpy as np
import pandas as pd

from sdgym._dataset_utils import (
    _get_dataset_subset,
    _parse_numeric_value,
    _read_metadata_json,
    _read_zipped_data,
)
from sdgym.s3 import (
    _list_s3_bucket_contents,
    _load_yaml_metainfo_from_s3,
    get_s3_client,
)

LOGGER = logging.getLogger(__name__)

DATASETS_PATH = Path(appdirs.user_data_dir()) / 'SDGym' / 'datasets'
BUCKET = 's3://sdv-datasets-public'
BUCKET_URL = 'https://{}.s3.amazonaws.com/'
TIMESERIES_FIELDS = ['sequence_index', 'entity_columns', 'context_columns', 'deepecho_version']
MODALITIES = ['single_table', 'multi_table', 'sequential']
S3_PREFIX = 's3://'


def _get_bucket_name(bucket):
    return bucket[len(S3_PREFIX) :] if bucket.startswith(S3_PREFIX) else bucket


def _download_dataset(
    modality,
    dataset_name,
    datasets_path=None,
    bucket=None,
    aws_access_key_id=None,
    aws_secret_access_key=None,
):
    """Download a dataset into the given ``datasets_path`` / ``modality``."""
    datasets_path = datasets_path or DATASETS_PATH / modality / dataset_name
    bucket = bucket or BUCKET
    bucket_name = _get_bucket_name(bucket)

    LOGGER.info('Downloading dataset %s from %s', dataset_name, bucket)
    s3_client = get_s3_client(aws_access_key_id, aws_secret_access_key)
    prefix = f'{modality.lower()}/{dataset_name}/'

    contents = _list_s3_bucket_contents(s3_client, bucket_name, prefix)
    if not contents:
        display_name = dataset_name
        if isinstance(dataset_name, Path):
            display_name = dataset_name.name

        raise ValueError(
            f"Dataset '{display_name}' not found in bucket '{bucket}' for modality '{modality}'."
        )

    for obj in contents:
        s3_path = obj['Key']
        local_path = datasets_path / os.path.relpath(s3_path, prefix)
        os.makedirs(os.path.dirname(local_path), exist_ok=True)
        s3_client.download_file(bucket_name, s3_path, local_path)

    return datasets_path


def _path_contains_data_and_metadata(dataset_path):
    metadata_found = False
    data_zip_found = False
    for file_name in dataset_path.iterdir():
        if 'metadata' in file_name.stem and file_name.suffix == '.json':
            metadata_found = True

        if 'data' in file_name.stem and file_name.suffix == '.zip':
            data_zip_found = True

    return metadata_found and data_zip_found


def _get_dataset_path_and_download(
    modality,
    dataset,
    datasets_path,
    bucket=None,
    aws_access_key_id=None,
    aws_secret_access_key=None,
):
    dataset = Path(dataset)
    if dataset.exists() and _path_contains_data_and_metadata(dataset):
        return dataset

    datasets_path = datasets_path or DATASETS_PATH / modality
    dataset_path = datasets_path / dataset
    if dataset_path.exists() and _path_contains_data_and_metadata(dataset_path):
        return dataset_path

    bucket = bucket or BUCKET
    if not bucket.startswith(S3_PREFIX):
        local_path = Path(bucket) / modality / dataset if bucket else Path(dataset)
        if local_path.exists() and _path_contains_data_and_metadata(local_path):
            return local_path

    dataset_path = _download_dataset(
        modality, dataset, dataset_path, bucket, aws_access_key_id, aws_secret_access_key
    )
    return dataset_path


def _validate_modality(modality):
    if modality not in MODALITIES:
        modalities_list = ', '.join(MODALITIES)
        raise ValueError(f'Modality `{modality}` not recognized. Must be one of {modalities_list}')


def get_data_and_metadata_from_path(dataset_path, modality):
    """Load dataset data and metadata from a given local path.

    Args:
        dataset_path (Path):
            The local path to the dataset directory containing data and metadata files.
        modality (str):
            The dataset modality. Used by the data-reading function to correctly
            interpret and load the dataset (e.g., ``'single_table'``, ``'multi_table'``).

    Returns:
        tuple[pd.DataFrame | dict, dict | None]:
            A tuple containing:
                - The loaded dataset as a ``pandas.DataFrame`` or a dictionary with
                  ``pandas.DataFrame`` if ``multi_table``.
            - The dataset metadata as a dictionary.
    """
    metadata_dict = None
    data = None
    for file_name in dataset_path.iterdir():
        if 'metadata' in file_name.stem and file_name.suffix == '.json':
            metadata_dict = _read_metadata_json(file_name)
        elif 'data' in file_name.stem and file_name.suffix == '.zip':
            data = _read_zipped_data(zip_file_path=(file_name), modality=modality)

        if data is not None and metadata_dict is not None:
            break

    return data, metadata_dict


def _genereate_dataset_info(s3_client, bucket_name, contents):
    """Generate summarized dataset information from S3 bucket.

    Args:
        s3_client (boto3.client):
            An initialized boto3 S3 client used to access S3 objects.
        bucket_name (str):
            The name of the S3 bucket containing the datasets.
        contents (list[dict]):
            A list of S3 object metadata dictionaries (e.g., from
            ``s3_client.list_objects_v2()['Contents']``). Each dictionary should
            include a ``'Key'`` field representing the S3 object path.

    Returns:
        dict:
            A dictionary containing dataset summary information with the
            following keys:
                - ``'dataset_name'`` (list[str]): Names of the datasets discovered.
                - ``'size_MB'`` (list[float]): Corresponding dataset sizes in megabytes.
                - ``'num_tables'`` (list[int]): Number of tables in each dataset.
    """
    tables_info = {'dataset_name': [], 'size_MB': [], 'num_tables': []}
    for obj in contents:
        key = obj.get('Key', '')
        if key.endswith('metainfo.yaml'):
            parts = key.split('/')
            if len(parts) >= 3:
                dataset_name = parts[-2]
                yaml_key = key
                info = _load_yaml_metainfo_from_s3(s3_client, bucket_name, yaml_key)
                size_mb = _parse_numeric_value(
                    info.get('dataset-size-mb', np.nan),
                    dataset_name,
                    field_name='dataset-size-mb',
                    target_type=float,
                )
                num_tables = _parse_numeric_value(
                    info.get('num-tables', np.nan), dataset_name, 'num-tables', target_type=int
                )
                tables_info['dataset_name'].append(dataset_name)
                tables_info['size_MB'].append(size_mb)
                tables_info['num_tables'].append(num_tables)

    return tables_info


def _get_available_datasets(
    modality, bucket=None, aws_access_key_id=None, aws_secret_access_key=None
):
    _validate_modality(modality)
    s3_client = get_s3_client(aws_access_key_id, aws_secret_access_key)
    bucket = bucket or BUCKET
    bucket_name = _get_bucket_name(bucket)
    contents = _list_s3_bucket_contents(s3_client, bucket_name, f'{modality}/')
    datasets_info = _genereate_dataset_info(s3_client, bucket_name, contents)
    return pd.DataFrame(datasets_info)


def load_dataset(
    modality,
    dataset,
    datasets_path=None,
    bucket=None,
    aws_access_key_id=None,
    aws_secret_access_key=None,
    limit_dataset_size=False,
):
    """Get the data and metadata of a dataset.

    Args:
        modality (str):
            It must be ``'single_table'``, ``'multi_table'`` or ``'sequential'``.
        dataset (str):
            The path of the dataset as a string or the name of the dataset.
        dataset_path (PurePath):
            The path of the dataset as an object. This will only be used if the given ``dataset``
            doesn't exist.
        bucket (str):
            The AWS bucket where to get the dataset. This will only be used if both ``dataset``
            and ``dataset_path`` don't exist.
        aws_access_key_id (str or None):
            The access key id that will be used to communicate with s3, if provided.
        aws_secret_access_key (str or None):
            The secret access key that will be used to communicate with s3, if provided.
        limit_dataset_size (bool):
            Use this flag to limit the size of the datasets for faster evaluation. If ``True``,
            limit the size of every table to 1,000 rows (randomly sampled) and the first 10
            columns. Defauts to ``False``.

    Returns:
        pd.DataFrame | dict, dict:
            The data and medatata for a dataset.
    """
    _validate_modality(modality)
    dataset_path = _get_dataset_path_and_download(
        modality, dataset, datasets_path, bucket, aws_access_key_id, aws_secret_access_key
    )

    data, metadata_dict = get_data_and_metadata_from_path(dataset_path, modality)
    if limit_dataset_size:
        data, metadata_dict = _get_dataset_subset(data, metadata_dict, modality=modality)

    return data, metadata_dict


def get_dataset_paths(
    modality,
    datasets=None,
    datasets_path=None,
    bucket=None,
    aws_access_key_id=None,
    aws_secret_access_key=None,
):
    """Build the full path to datasets and ensure they exist.

    Args:
        datasets (list):
            List of datasets.
        dataset_path (str):
            The path of the datasets.
        bucket (str):
            The AWS bucket where to get the dataset or folder.
        aws_access_key_id (str):
            The access key id that will be used to communicate with s3, if provided.
        aws_secret_access_key (str):
            The secret access key that will be used to communicate with s3, if provided.

    Returns:
        list:
            List of the full path of the datasets.
    """
    _validate_modality(modality)
    bucket = bucket or BUCKET
    is_remote = bucket.startswith(S3_PREFIX)

    if datasets_path is None:
        datasets_path = DATASETS_PATH / modality
    else:
        datasets_path = Path(datasets_path)

    if datasets is None:
        if not is_remote and Path(bucket).exists():
            datasets = []
            folder_items = list(Path(bucket).iterdir())
            for dataset in folder_items:
                if _path_contains_data_and_metadata(dataset) and dataset not in datasets:
                    datasets.append(dataset)
        else:
            datasets = _get_available_datasets(
                modality,
                bucket=bucket,
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
            )
            datasets = datasets['dataset_name'].tolist()

    dataset_paths = []
    for dataset in datasets:
        available_dataset = _get_dataset_path_and_download(
            modality, dataset, datasets_path, bucket, aws_access_key_id, aws_secret_access_key
        )
        dataset_paths.append(available_dataset)

    return dataset_paths



"""Random utils used by SDGym."""

import logging
import os
import subprocess
import sys
import traceback
from datetime import datetime, timezone

import humanfriendly
import numpy as np
import pandas as pd
import psutil

from sdgym.errors import SDGymError
from sdgym.synthesizers.base import BaselineSynthesizer
from sdgym.synthesizers.sdv import _get_all_sdv_synthesizers, create_sdv_synthesizer_class

LOGGER = logging.getLogger(__name__)


def used_memory():
    """Get the memory used by this process nicely formatted."""
    process = psutil.Process(os.getpid())
    return humanfriendly.format_size(process.memory_info().rss)


def format_exception():
    """Format exceptions."""
    exception = traceback.format_exc()
    exc_type, exc_value, _ = sys.exc_info()
    error = traceback.format_exception_only(exc_type, exc_value)[0].strip()
    return exception, error


def get_synthesizers(synthesizers):
    """Get the dict of synthesizer name and object for each synthesizer.

    Args:
        synthesizers (list):
            An iterable of synthesizer classes and strings.

    Returns:
        dict[str, function]:
            Dict with the synthesizer name and object.

    Raises:
        TypeError:
            If neither a list is not passed.
    """
    synthesizers = [] if synthesizers is None else synthesizers
    if not isinstance(synthesizers, list):
        raise TypeError('`synthesizers` must be a list.')

    synthesizers_dicts = []
    baselines = BaselineSynthesizer.get_subclasses(include_parents=True)
    for synthesizer in synthesizers:
        if isinstance(synthesizer, str):
            LOGGER.info('Trying to import synthesizer by name.')
            synthesizer_class = baselines.get(synthesizer)
            if synthesizer_class is None:
                if synthesizer in _get_all_sdv_synthesizers():
                    synthesizer_class = create_sdv_synthesizer_class(synthesizer)
                else:
                    raise SDGymError(f'Unknown synthesizer {synthesizer}') from None

            synthesizers_dicts.append({
                'name': synthesizer,
                'synthesizer': synthesizer_class(),
            })
            continue

        if isinstance(synthesizer, type) or hasattr(synthesizer, '__name__'):
            synthesizer_name = getattr(synthesizer, '__name__', 'undefined')
        else:
            synthesizer_name = getattr(type(synthesizer), '__name__', 'undefined')
        synthesizers_dicts.append({
            'name': synthesizer_name,
            'synthesizer': synthesizer,
        })

    return synthesizers_dicts


def get_size_of(obj, obj_ids=None):
    """Get the memory used by a given object in bytes.

    Args:
        obj (object):
            The object to get the size of.
        obj_ids (set):
            The ids of the objects that have already been evaluated.

    Returns:
        int:
            The size in bytes.
    """
    size = 0
    if obj_ids is None:
        obj_ids = set()

    obj_id = id(obj)
    if obj_id in obj_ids:
        return 0

    obj_ids.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size_of(v, obj_ids) for v in obj.values()])
    elif isinstance(obj, pd.DataFrame):
        size += obj.memory_usage(index=True).sum()
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size_of(i, obj_ids) for i in obj])
    else:
        size += sys.getsizeof(obj)

    return size


def get_duplicates(items):
    """Get any duplicate items in the given list.

    Args:
        items (list):
            The list of items to de-duplicate.

    Returns:
        set:
            The duplicate items.
    """
    seen = set()
    return {item for item in items if item in seen or seen.add(item)}


def get_num_gpus():
    """Get number of gpus.

    Returns:
        int:
            Number of gpus to use.
    """
    try:
        command = ['nvidia-smi', '--query-gpu=utilization.gpu', '--format=csv,noheader,nounits']
        output = subprocess.run(command, stdout=subprocess.PIPE)
        return len(output.stdout.decode().split())

    except Exception:
        return 0


def select_device():
    """Select gpu if available, otherwise select cpu.

    Returns:
        str:
            The cuda device if available, otherwise ``'cpu'``.
    """
    try:
        command = ['nvidia-smi', '--query-gpu=utilization.gpu', '--format=csv,noheader,nounits']
        output = subprocess.run(command, stdout=subprocess.PIPE)
        loads = np.array(output.stdout.decode().split()).astype(float)
        device = loads.argmin()
        return f'cuda:{device}'

    except Exception:
        return 'cpu'


def get_utc_now():
    """Get the current datetime, in UTC timezone.

    Returns:
        datetime:
            The current datetime in UTC.
    """
    return datetime.now(tz=timezone.utc)


def calculate_score_time(start):
    """Calculate the total duration in seconds.

    Args:
        start (datetime):
            The start datetime to use to calculate duration.

    Returns:
        float:
            The duration in total seconds.

    """
    return (get_utc_now() - start).total_seconds()


def convert_metadata_to_sdmetrics(metadata_dict):
    """Convert a sdv metadata dictionary into sdmetrics expected metadata."""
    table_name = next(iter(metadata_dict['tables']))
    return metadata_dict['tables'][table_name]



"""ColumnSynthesizer module."""

import logging

import pandas as pd
from rdt.hyper_transformer import HyperTransformer
from sdv.metadata import Metadata
from sklearn.mixture import GaussianMixture

from sdgym.synthesizers.base import BaselineSynthesizer

LOGGER = logging.getLogger(__name__)


class ColumnSynthesizer(BaselineSynthesizer):
    """Synthesizer that learns each column independently.

    Categorical columns are sampled using empirical frequencies.
    Continuous columns are learned and sampled using a GMM.
    """

    def _get_trained_synthesizer(self, real_data, metadata):
        hyper_transformer = HyperTransformer()
        hyper_transformer.detect_initial_config(real_data)
        supported_sdtypes = hyper_transformer._get_supported_sdtypes()
        config = {}
        if isinstance(metadata, Metadata):
            table_name = metadata._get_single_table_name()
            columns = metadata.tables[table_name].columns
        else:
            columns = metadata.columns

        for column_name, column in columns.items():
            sdtype = column['sdtype']
            if sdtype in supported_sdtypes:
                config[column_name] = sdtype
            elif column.get('pii', False):
                config[column_name] = 'pii'
            else:
                LOGGER.info(
                    f'Column {column} sdtype: {sdtype} is not supported, '
                    f'defaulting to inferred type.'
                )

        hyper_transformer.update_sdtypes(config)

        # This is done to match the behavior of the synthesizer for SDGym <= 0.6.0
        columns_to_remove = [
            column_name for column_name, data in real_data.items() if data.dtype.kind in {'O', 'i'}
        ]
        hyper_transformer.remove_transformers(columns_to_remove)

        hyper_transformer.fit(real_data)
        transformed = hyper_transformer.transform(real_data)

        self.length = len(real_data)
        gm_models = {}
        for name, column in transformed.items():
            kind = column.dtype.kind
            if kind != 'O':
                num_components = min(column.nunique(), 5)
                model = GaussianMixture(num_components)
                model.fit(column.to_numpy().reshape(-1, 1))
                gm_models[name] = model

        return (hyper_transformer, transformed, gm_models)

    def _sample_from_synthesizer(self, synthesizer, n_samples):
        hyper_transformer, transformed, gm_models = synthesizer
        sampled = pd.DataFrame()
        for name, column in transformed.items():
            kind = column.dtype.kind
            if kind == 'O':
                values = column.sample(self.length, replace=True).to_numpy()
            else:
                model = gm_models.get(name)
                values = model.sample(self.length)[0].ravel().clip(column.min(), column.max())

            sampled[name] = values

        return hyper_transformer.reverse_transform(sampled)



"""Metrics module."""

import sdmetrics


class WithKWargs:
    """Wrapper for sdmetrics.

    Args:
        metric (sdmetric):
            Metric object from sdmetrics.
        kwargs (dict):
            Key word arguments to use for the metric.
    """

    def __init__(self, metric, **kwargs):
        self._metric = metric
        self._kwargs = kwargs

    def compute(self, real_data, synthetic_data, metadata):
        """Compute the metric.

        Args:
            real_data (pandas.DataFrame):
                The values from the real dataset, passed as a pandas.DataFrame.
            synthetic_data (pandas.DataFrame):
                The values from the synthetic dataset, passed as a pandas.DataFrame.
            metadata (dict):
                Metadata dict. If ``None``, it is build based on the real_data fields and dtypes.

        Returns:
            Union[float, tuple[float]]:
                Metric output.
        """
        return self._metric.compute(real_data, synthetic_data, metadata, **self._kwargs)

    def normalize(self, raw_score):
        """Normalize the metric.

        Args:
            raw_score (float):
                The score.

        Returns:
            float:
                The normalized score.
        """
        return self._metric.normalize(raw_score)


# Metrics to use by default for specific problem types and data
# modalities if no metrics have been explicitly specified.
PROBLEM_TYPE_METRICS = {
    'binary_classification': [
        'BinaryDecisionTreeClassifier',
        'BinaryAdaBoostClassifier',
        'BinaryLogisticRegression',
        'BinaryMLPClassifier',
    ],
    'multiclass_classification': [
        'MulticlassDecisionTreeClassifier',
        'MulticlassMLPClassifier',
    ],
    'regression': [
        'LinearRegression',
        'MLPRegressor',
    ],
    'bayesian_likelihood': [
        'BNLogLikelihood',
    ],
    'gaussian_likelihood': [
        (
            'GMLogLikelihood(10)',
            WithKWargs(sdmetrics.single_table.GMLogLikelihood, n_components=10, iterations=10),
        ),
        (
            'GMLogLikelihood(30)',
            WithKWargs(sdmetrics.single_table.GMLogLikelihood, n_components=30, iterations=10),
        ),
    ],
}
DATA_MODALITY_METRICS = {
    'single-table': [
        'CSTest',
        'KSComplement',
    ],
    'multi-table': [
        'CSTest',
        'KSComplement',
    ],
    'timeseries': [
        'TSFClassifierEfficacy',
        'LSTMClassifierEfficacy',
        'TSFCDetection',
        'LSTMDetection',
    ],
}


def get_metrics(metrics, modality):
    """Get metrics for a given modality.

    Args:
        metrics (list):
            List of strings or tuples ``(metric, metric_args)`` describing the metrics.
        modality (str):
            It must be ``'single-table'``, ``'multi-table'`` or ``'timeseries'``.

    Returns:
        list, kwargs:
            A list of metrics for the given modality, and their corresponding kwargs.
    """
    if modality == 'multi-table':
        metric_classes = sdmetrics.multi_table.MultiTableMetric.get_subclasses()
    elif modality == 'single-table':
        metric_classes = sdmetrics.single_table.SingleTableMetric.get_subclasses()
    elif modality == 'timeseries':
        metric_classes = sdmetrics.timeseries.TimeSeriesMetric.get_subclasses()

    if not metrics:
        metrics = DATA_MODALITY_METRICS[modality]

    final_metrics = {}
    metric_kwargs = {}
    for metric in metrics:
        metric_args = None
        if isinstance(metric, tuple):
            metric, metric_args = metric
        if isinstance(metric, str):
            metric_name = metric
            try:
                metric = metric_classes[metric]
            except KeyError:
                raise ValueError(f'Unknown {modality} metric: {metric}') from None

        else:
            metric_name = metric.__name__

        final_metrics[metric_name] = metric
        if metric_args:
            metric_kwargs[metric_name] = metric_args

    return final_metrics, metric_kwargs



"""Synthesizers module."""

from sdgym.synthesizers.generate import (
    create_synthesizer_variant,
    create_single_table_synthesizer,
    create_multi_table_synthesizer,
)
from sdgym.synthesizers.identity import DataIdentity
from sdgym.synthesizers.column import ColumnSynthesizer
from sdgym.synthesizers.realtabformer import RealTabFormerSynthesizer
from sdgym.synthesizers.uniform import UniformSynthesizer
from sdgym.synthesizers.utils import (
    get_available_single_table_synthesizers,
    get_available_multi_table_synthesizers,
)
from sdgym.synthesizers.sdv import create_sdv_synthesizer_class, _get_all_sdv_synthesizers


__all__ = [
    'DataIdentity',
    'ColumnSynthesizer',
    'UniformSynthesizer',
    'RealTabFormerSynthesizer',
    'create_single_table_synthesizer',
    'create_multi_table_synthesizer',
    'create_synthesizer_variant',
    'get_available_single_table_synthesizers',
    'get_available_multi_table_synthesizers',
]

for sdv_name in _get_all_sdv_synthesizers():
    create_sdv_synthesizer_class(sdv_name)



"""SDGym Results Explorer for accessing and managing benchmark results."""

import os

from sdgym.benchmark import DEFAULT_DATASETS
from sdgym.datasets import load_dataset
from sdgym.result_explorer.result_handler import LocalResultsHandler, S3ResultsHandler
from sdgym.s3 import _get_s3_client, is_s3_path


def _validate_local_path(path):
    """Validates that the local path exists and is a directory."""
    if not os.path.isdir(path):
        raise ValueError(f"The provided path '{path}' is not a valid local directory.")


class ResultsExplorer:
    """Explorer for SDGym benchmark results, supporting both local and S3 storage."""

    def __init__(self, path, aws_access_key_id=None, aws_secret_access_key=None):
        self.path = path
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key

        if is_s3_path(path):
            s3_client = _get_s3_client(path, aws_access_key_id, aws_secret_access_key)
            self._handler = S3ResultsHandler(path, s3_client)
        else:
            _validate_local_path(path)
            self._handler = LocalResultsHandler(path)

    def list(self):
        """List all runs available in the results directory."""
        return self._handler.list()

    def _get_file_path(self, results_folder_name, dataset_name, synthesizer_name, file_type):
        """Validate access to the synthesizer or synthetic data file."""
        end_filename = f'{synthesizer_name}'
        if file_type == 'synthetic_data':
            end_filename += '_synthetic_data.csv'
        elif file_type == 'synthesizer':
            end_filename += '.pkl'

        date = '_'.join(results_folder_name.split('_')[-3:])
        path_parts = [results_folder_name, f'{dataset_name}_{date}', synthesizer_name]

        return self._handler.get_file_path(path_parts, end_filename)

    def load_synthesizer(self, results_folder_name, dataset_name, synthesizer_name):
        """Load the synthesizer for a given dataset and synthesizer."""
        file_path = self._get_file_path(
            results_folder_name, dataset_name, synthesizer_name, 'synthesizer'
        )
        return self._handler.load_synthesizer(file_path)

    def load_synthetic_data(self, results_folder_name, dataset_name, synthesizer_name):
        """Load the synthetic data for a given dataset and synthesizer."""
        file_path = self._get_file_path(
            results_folder_name, dataset_name, synthesizer_name, 'synthetic_data'
        )
        return self._handler.load_synthetic_data(file_path)

    def load_real_data(self, dataset_name):
        """Load the real data for a given dataset."""
        if dataset_name not in DEFAULT_DATASETS:
            raise ValueError(
                f"Dataset '{dataset_name}' is not a SDGym dataset. "
                'Please provide a valid dataset name.'
            )

        data, _ = load_dataset(
            modality='single_table',
            dataset=dataset_name,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
        )
        return data

    def summarize(self, folder_name):
        """Summarize the results in the specified folder.

        Args:
            folder_name (str):
                The name of the results folder to summarize.

        Returns:
            tuple (pd.DataFrame, pd.DataFrame):
                - A summary DataFrame with the number of Wins per synthesizer.
                - A DataFrame with the results of the benchmark for the specified folder.
        """
        return self._handler.summarize(folder_name)

    def all_runs_complete(self, folder_name):
        """Check if all runs in the specified folder are complete."""
        return self._handler.all_runs_complete(folder_name)

    def load_results(self, results_folder_name):
        """Load and aggregate all the results CSV files from the specified results folder.

        Args:
            results_folder_name (str):
                The name of the results folder to load results from.

        Returns:
            pd.DataFrame:
                A DataFrame containing the results of the specified folder.
        """
        return self._handler.load_results(results_folder_name)

    def load_metainfo(self, results_folder_name):
        """Load and aggregate all the metainfo YAML files from the specified results folder.

        Args:
            results_folder_name (str):
                The name of the results folder to load metainfo from.

        Returns:
            dict:
                A dictionary containing the metainfo of the specified folder.
        """
        return self._handler.load_metainfo(results_folder_name)



"""Tests for the ``tasks.py`` file."""

from tasks import _get_extra_dependencies, _get_minimum_versions, _resolve_version_conflicts


def test_get_minimum_versions():
    """Test the ``_get_minimum_versions`` method.

    The method should return the minimum versions of the dependencies for the given python version.
    If a library is linked to an URL, the minimum version should be the URL.
    """
    # Setup
    dependencies = [
        "numpy>=1.20.0,<2;python_version<'3.10'",
        "numpy>=1.23.3,<2;python_version>='3.10'",
        "pandas>=1.2.0,<2;python_version<'3.10'",
        "pandas>=1.3.0,<2;python_version>='3.10'",
        'humanfriendly>=8.2,<11',
        'pandas @ git+https://github.com/pandas-dev/pandas.git@master',
    ]

    # Run
    minimum_versions_39 = _get_minimum_versions(dependencies, '3.9')
    minimum_versions_310 = _get_minimum_versions(dependencies, '3.10')

    # Assert
    expected_versions_39 = {
        'numpy': 'numpy==1.20.0',
        'pandas': 'git+https://github.com/pandas-dev/pandas.git@master#egg=pandas',
        'humanfriendly': 'humanfriendly==8.2',
    }
    expected_versions_310 = {
        'numpy': 'numpy==1.23.3',
        'pandas': 'git+https://github.com/pandas-dev/pandas.git@master#egg=pandas',
        'humanfriendly': 'humanfriendly==8.2',
    }

    assert minimum_versions_39 == expected_versions_39
    assert minimum_versions_310 == expected_versions_310


def _get_example_pyproject_dict():
    return {
        'build-system': {
            'build-backend': 'setuptools.build_meta',
            'requires': ['setuptools', 'wheel'],
        },
        'project': {
            'authors': [{'email': 'info@sdv.dev', 'name': 'DataCebo, Inc.'}],
            'classifiers': [
                'Intended Audience :: Developers',
                'License :: Free for non-commercial use',
                'Natural Language :: English',
                'Programming Language :: Python :: 3.10',
                'Programming Language :: Python :: 3.11',
                'Programming Language :: Python :: 3.12',
                'Topic :: Scientific/Engineering :: Artificial Intelligence',
            ],
            'dependencies': [
                'appdirs>=1.3',
                'boto3>=1.28,<2',
                'botocore>=1.31,<2',
                'cloudpickle>=2.1.0',
                'compress-pickle>=1.2.0',
                'humanfriendly>=8.2',
                "numpy>=1.21.6;python_version<'3.10'",
                "numpy>=1.23.3;python_version>='3.10' and python_version<'3.12'",
                "numpy>=1.26.0;python_version>='3.12'",
                "pandas>=1.4.0;python_version<'3.11'",
                "pandas>=1.5.0;python_version>='3.11' and python_version<'3.12'",
                "pandas>=2.1.1;python_version>='3.12'",
                'psutil>=5.7',
                "scikit-learn>=1.0.2;python_version<'3.10'",
                "scikit-learn>=1.1.0;python_version>='3.10' and python_version<'3.11'",
                "scikit-learn>=1.1.3;python_version>='3.11' and python_version<'3.12'",
                "scikit-learn>=1.3.1;python_version>='3.12'",
                "scipy>=1.7.3;python_version<'3.10'",
                "scipy>=1.9.2;python_version>='3.10' and python_version<'3.12'",
                "scipy>=1.12.0;python_version>='3.12'",
                'tabulate>=0.8.3,<0.9',
                "torch>=1.12.1;python_version<'3.10'",
                "torch>=2.0.0;python_version>='3.10' and python_version<'3.12'",
                "torch>=2.2.0;python_version>='3.12'",
                'tqdm>=4.66.3',
                'XlsxWriter>=1.2.8',
                'rdt>=1.13.1',
                'sdmetrics>=0.17.0',
                'sdv>=1.17.2',
            ],
            'dynamic': ['version'],
            'license': {'text': 'BSL-1.1'},
            'name': 'sdgym',
            'optional-dependencies': {
                'all': ['sdgym[dask, test, dev]'],
                'dask': ['dask', 'distributed'],
                'dev': [
                    'sdgym[dask, test]',
                    'build>=1.0.0,<2',
                    'bump-my-version>=0.18.3,<1',
                    'pip>=9.0.1',
                    'watchdog>=1.0.1,<5',
                    'ruff>=0.4.5,<1',
                    'twine>=1.10.0,<6',
                    'wheel>=0.30.0',
                    'coverage>=4.5.12,<8',
                    'tox>=2.9.1,<5',
                    'importlib-metadata>=3.6',
                    'invoke',
                ],
                'realtabformer': ['realtabformer>=0.2.1'],
                'test': [
                    'sdgym[realtabformer]',
                    'pytest>=6.2.5',
                ],
            },
            'readme': 'README.md',
            'requires-python': '>=3.9,<3.13',
        },
        'tool': {
            'bumpversion': {
                'allow_dirty': False,
                'commit': True,
                'commit_args': '',
            },
            'ruff': {
                'exclude': [
                    'docs',
                    '.tox',
                    '.git',
                    '__pycache__',
                    '.ipynb_checkpoints',
                    'tasks.py',
                ],
                'indent-width': 4,
            },
        },
    }


def test__get_extra_dependencies():
    """Test that the proper dependency strings are extracted from the pyproject dictionary."""
    # Setup
    pyproject_dict = _get_example_pyproject_dict()

    # Run
    extra_dependencies = _get_extra_dependencies(pyproject_dict)

    # Assert
    assert extra_dependencies == ['realtabformer>=0.2.1']


def test__resolve_version_conflicts_conflicting_versions():
    """Test that any conflicts for the same dependency are resolved to the higher version."""
    # Setup
    deps = {
        'numpy': 'numpy==2.0.1',
        'pandas': 'pandas==2.2.1',
        'sdv': 'sdv==2.1.1',
        'rdt': 'rdt==1.1.2',
    }
    extra_deps = {
        'numpy': 'numpy==2.0.0',
        'pandas': 'pandas==2.3.0',
        'sdv': 'sdv==3.0.0',
        'copulas': 'copulas==0.12.0',
    }

    # Run
    versions = _resolve_version_conflicts(deps, extra_deps)

    # Assert
    assert sorted(versions) == sorted([
        'numpy==2.0.1',
        'pandas==2.3.0',
        'sdv==3.0.0',
        'rdt==1.1.2',
        'copulas==0.12.0',
    ])


def test__resolve_version_conflicts_pointing_to_branch():
    """Test specific branches are always selected over normal version numbers."""
    # Setup
    deps = {
        'numpy': 'git+https://github.com/numpy-dev/numpy.git@master#egg=numpy',
        'pandas': 'pandas==2.2.1',
        'sdv': 'sdv==2.1.1',
        'rdt': 'rdt==1.1.2',
    }
    extra_deps = {
        'numpy': 'numpy==2.0.0',
        'pandas': 'git+https://github.com/pandas-dev/pandas.git@master#egg=pandas',
        'sdv': 'sdv==3.0.0',
        'copulas': 'copulas==0.12.0',
    }

    # Run
    versions = _resolve_version_conflicts(deps, extra_deps)

    # Assert
    assert sorted(versions) == sorted([
        'git+https://github.com/numpy-dev/numpy.git@master#egg=numpy',
        'git+https://github.com/pandas-dev/pandas.git@master#egg=pandas',
        'sdv==3.0.0',
        'rdt==1.1.2',
        'copulas==0.12.0',
    ])



"""Base classes for synthesizers."""

import abc
import logging
import warnings

from sdv.metadata import Metadata

LOGGER = logging.getLogger(__name__)


class BaselineSynthesizer(abc.ABC):
    """Base class for all the ``SDGym`` baselines."""

    _MODEL_KWARGS = {}
    _NATIVELY_SUPPORTED = True

    @classmethod
    def get_subclasses(cls, include_parents=False):
        """Recursively find subclasses of this Baseline.

        Args:
            include_parents (bool):
                Whether to include subclasses which are parents to
                other classes. Defaults to ``False``.
        """
        subclasses = {}
        for child in cls.__subclasses__():
            grandchildren = child.get_subclasses(include_parents)
            subclasses.update(grandchildren)
            if include_parents or not grandchildren:
                subclasses[child.__name__] = child

        return subclasses

    @classmethod
    def _get_supported_synthesizers(cls):
        """Get the natively supported synthesizer class names."""
        subclasses = cls.get_subclasses(include_parents=True)
        synthesizers = set()
        for name, subclass in subclasses.items():
            if subclass._NATIVELY_SUPPORTED:
                synthesizers.add(name)

        return sorted(synthesizers)

    @classmethod
    def get_baselines(cls):
        """Get baseline classes."""
        subclasses = cls.get_subclasses(include_parents=True)
        synthesizers = []
        for _, subclass in subclasses.items():
            if abc.ABC not in subclass.__bases__:
                synthesizers.append(subclass)

        return synthesizers

    def get_trained_synthesizer(self, data, metadata):
        """Get a synthesizer that has been trained on the provided data and metadata.

        Args:
            data (pandas.DataFrame):
                The data to train on.
            metadata (dict):
                The metadata dictionary.

        Returns:
            obj:
                The synthesizer object.
        """
        metadata_object = Metadata()
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', UserWarning)
            metadata = metadata_object.load_from_dict(metadata)

        return self._get_trained_synthesizer(data, metadata)

    def sample_from_synthesizer(self, synthesizer, n_samples):
        """Sample data from the provided synthesizer.

        Args:
            synthesizer (obj):
                The synthesizer object to sample data from.
            n_samples (int):
                The number of samples to create.

        Returns:
            pandas.DataFrame or dict:
                The sampled data. If single-table, should be a DataFrame. If multi-table,
                should be a dict mapping table name to DataFrame.
        """
        return self._sample_from_synthesizer(synthesizer, n_samples)



"""Main SDGym benchmarking module."""

import concurrent
import logging
import math
import multiprocessing
import os
import pickle
import re
import textwrap
import threading
import tracemalloc
import warnings
from collections import defaultdict
from contextlib import contextmanager
from datetime import datetime
from importlib.metadata import version
from pathlib import Path
from urllib.parse import urlparse

import boto3
import cloudpickle
import compress_pickle
import numpy as np
import pandas as pd
import tqdm
from botocore.config import Config
from sdmetrics.reports.multi_table import (
    DiagnosticReport as MultiTableDiagnosticReport,
)
from sdmetrics.reports.multi_table import (
    QualityReport as MultiTableQualityReport,
)
from sdmetrics.reports.single_table import (
    DiagnosticReport as SingleTableDiagnosticReport,
)
from sdmetrics.reports.single_table import (
    QualityReport as SingleTableQualityReport,
)
from sdmetrics.single_table import DCRBaselineProtection

from sdgym.datasets import get_dataset_paths, load_dataset
from sdgym.errors import SDGymError
from sdgym.metrics import get_metrics
from sdgym.progress import TqdmLogger, progress
from sdgym.result_writer import LocalResultsWriter, S3ResultsWriter
from sdgym.s3 import (
    S3_PREFIX,
    S3_REGION,
    is_s3_path,
    parse_s3_path,
    write_csv,
    write_file,
)
from sdgym.synthesizers import UniformSynthesizer
from sdgym.synthesizers.base import BaselineSynthesizer
from sdgym.utils import (
    calculate_score_time,
    convert_metadata_to_sdmetrics,
    format_exception,
    get_duplicates,
    get_num_gpus,
    get_size_of,
    get_synthesizers,
    get_utc_now,
    used_memory,
)

LOGGER = logging.getLogger(__name__)
DEFAULT_SYNTHESIZERS = ['GaussianCopulaSynthesizer', 'CTGANSynthesizer', 'UniformSynthesizer']
DEFAULT_DATASETS = [
    'adult',
    'alarm',
    'census',
    'child',
    'covtype',
    'expedia_hotel_logs',
    'insurance',
    'intrusion',
    'news',
]
N_BYTES_IN_MB = 1000 * 1000
EXTERNAL_SYNTHESIZER_TO_LIBRARY = {
    'RealTabFormerSynthesizer': 'realtabformer',
}
FILE_INCREMENT_PATTERN = re.compile(r'\((\d+)\)$')
RESULTS_DATE_PATTERN = re.compile(r'SDGym_results_(\d{2}_\d{2}_\d{4})')
METAINFO_FILE_PATTERN = re.compile(r'metainfo(?:\((\d+)\))?\.yaml$')
SDV_SINGLE_TABLE_SYNTHESIZERS = [
    'GaussianCopulaSynthesizer',
    'CTGANSynthesizer',
    'CopulaGANSynthesizer',
    'TVAESynthesizer',
]


def _validate_inputs(output_filepath, detailed_results_folder, synthesizers, custom_synthesizers):
    if output_filepath and os.path.exists(output_filepath):
        raise ValueError(
            f'{output_filepath} already exists. Please provide a file that does not already exist.'
        )

    if detailed_results_folder and os.path.exists(detailed_results_folder):
        raise ValueError(
            f'{detailed_results_folder} already exists. '
            'Please provide a folder that does not already exist.'
        )

    duplicates = get_duplicates(synthesizers) if synthesizers else {}
    if custom_synthesizers:
        duplicates.update(get_duplicates(custom_synthesizers))
    if len(duplicates) > 0:
        raise ValueError(
            'Synthesizers must be unique. Please remove repeated values in the `synthesizers` '
            'and `custom_synthesizers` parameters.'
        )


def _create_detailed_results_directory(detailed_results_folder):
    if detailed_results_folder and not is_s3_path(detailed_results_folder):
        detailed_results_folder = Path(detailed_results_folder)
        os.makedirs(detailed_results_folder, exist_ok=True)


def _get_metainfo_increment(top_folder, s3_client=None):
    increments = []
    first_file_message = 'No metainfo file found, starting from increment (0)'
    if s3_client:
        bucket, prefix = parse_s3_path(top_folder)
        try:
            response = s3_client.list_objects_v2(Bucket=bucket, Prefix=prefix)
            contents = response.get('Contents', [])
            for obj in contents:
                file_name = Path(obj['Key']).name
                match = METAINFO_FILE_PATTERN.match(file_name)
                if match:
                    # Extract numeric suffix (e.g. metainfo(3).yaml → 3) or 0 if plain metainfo.yaml
                    increments.append(int(match.group(1)) if match.group(1) else 0)
        except Exception:
            LOGGER.info(first_file_message)
            return 0  # start with (0) if error
    else:
        top_folder = Path(top_folder)
        if not top_folder.exists():
            LOGGER.info(first_file_message)
            return 0
        for file in top_folder.glob('metainfo*.yaml'):
            match = METAINFO_FILE_PATTERN.match(file.name)
            if match:
                increments.append(int(match.group(1)) if match.group(1) else 0)

    return max(increments) + 1 if increments else 0


def _setup_output_destination_aws(output_destination, synthesizers, datasets, s3_client):
    paths = defaultdict(dict)
    s3_path = output_destination[len(S3_PREFIX) :].rstrip('/')
    parts = s3_path.split('/')
    bucket_name = parts[0]
    prefix_parts = parts[1:]
    paths['bucket_name'] = bucket_name
    today = datetime.today().strftime('%m_%d_%Y')
    top_folder = '/'.join(prefix_parts + [f'SDGym_results_{today}'])
    increment = _get_metainfo_increment(f's3://{bucket_name}/{top_folder}', s3_client)
    suffix = f'({increment})' if increment >= 1 else ''
    s3_client.put_object(Bucket=bucket_name, Key=top_folder + '/')
    for dataset in datasets:
        dataset_folder = f'{top_folder}/{dataset}_{today}'
        s3_client.put_object(Bucket=bucket_name, Key=dataset_folder + '/')
        paths[dataset]['meta'] = f's3://{bucket_name}/{dataset_folder}/meta.yaml'
        for synth_name in synthesizers:
            final_synth_name = f'{synth_name}{suffix}'
            synth_folder = f'{dataset_folder}/{final_synth_name}'
            s3_client.put_object(Bucket=bucket_name, Key=synth_folder + '/')
            paths[dataset][final_synth_name] = {
                'synthesizer': f's3://{bucket_name}/{synth_folder}/{final_synth_name}.pkl',
                'synthetic_data': f's3://{bucket_name}/{synth_folder}/{final_synth_name}_synthetic_data.csv',
                'benchmark_result': f's3://{bucket_name}/{synth_folder}/{final_synth_name}_benchmark_result.csv',
                'results': f's3://{bucket_name}/{top_folder}/results{suffix}.csv',
                'metainfo': f's3://{bucket_name}/{top_folder}/metainfo{suffix}.yaml',
            }

    s3_client.put_object(
        Bucket=bucket_name,
        Key=f'{top_folder}/metainfo{suffix}.yaml',
        Body='completed_date: null\n'.encode('utf-8'),
    )
    return paths


def _setup_output_destination(output_destination, synthesizers, datasets, s3_client=None):
    """Set up the output destination for the benchmark results.

    Args:
        output_destination (str or None):
            The path to the output directory where results will be saved.
            If None, no output will be saved.
        synthesizers (list):
            The list of synthesizers to benchmark.
        datasets (list):
            The list of datasets to benchmark.
    """
    if s3_client:
        return _setup_output_destination_aws(output_destination, synthesizers, datasets, s3_client)

    if output_destination is None:
        return {}

    output_path = Path(output_destination)
    output_path.mkdir(parents=True, exist_ok=True)
    today = datetime.today().strftime('%m_%d_%Y')
    top_folder = output_path / f'SDGym_results_{today}'
    top_folder.mkdir(parents=True, exist_ok=True)
    increment = _get_metainfo_increment(top_folder)
    suffix = f'({increment})' if increment >= 1 else ''
    paths = defaultdict(dict)
    for dataset in datasets:
        dataset_folder = top_folder / f'{dataset}_{today}'
        dataset_folder.mkdir(parents=True, exist_ok=True)

        for synth_name in synthesizers:
            final_synth_name = f'{synth_name}{suffix}'
            synth_folder = dataset_folder / final_synth_name
            synth_folder.mkdir(parents=True, exist_ok=True)
            paths[dataset][final_synth_name] = {
                'synthesizer': str(synth_folder / f'{final_synth_name}.pkl'),
                'synthetic_data': str(synth_folder / f'{final_synth_name}_synthetic_data.csv'),
                'benchmark_result': str(synth_folder / f'{final_synth_name}_benchmark_result.csv'),
                'metainfo': str(top_folder / f'metainfo{suffix}.yaml'),
                'results': str(top_folder / f'results{suffix}.csv'),
            }

    return paths


def _generate_job_args_list(
    limit_dataset_size,
    sdv_datasets,
    additional_datasets_folder,
    sdmetrics,
    detailed_results_folder,
    timeout,
    output_destination,
    compute_quality_score,
    compute_diagnostic_score,
    compute_privacy_score,
    synthesizers,
    custom_synthesizers,
    s3_client,
):
    # Get list of synthesizer objects
    synthesizers = [] if synthesizers is None else synthesizers
    custom_synthesizers = [] if custom_synthesizers is None else custom_synthesizers
    synthesizers = get_synthesizers(synthesizers + custom_synthesizers)

    # Get list of dataset paths
    aws_access_key_id = os.getenv('AWS_ACCESS_KEY_ID')
    aws_secret_access_key_key = os.getenv('AWS_SECRET_ACCESS_KEY')
    sdv_datasets = (
        []
        if sdv_datasets is None
        else get_dataset_paths(
            modality='single_table',
            datasets=sdv_datasets,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key_key,
        )
    )
    additional_datasets = (
        []
        if additional_datasets_folder is None
        else get_dataset_paths(
            modality='single_table',
            bucket=(
                additional_datasets_folder
                if is_s3_path(additional_datasets_folder)
                else os.path.join(additional_datasets_folder, 'single_table')
            ),
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key_key,
        )
    )
    datasets = sdv_datasets + additional_datasets
    synthesizer_names = [synthesizer['name'] for synthesizer in synthesizers]
    dataset_names = [dataset.name for dataset in datasets]
    paths = _setup_output_destination(
        output_destination, synthesizer_names, dataset_names, s3_client=s3_client
    )
    job_tuples = []
    for dataset in datasets:
        for synthesizer in synthesizers:
            if paths:
                final_name = next(
                    (name for name in paths[dataset.name] if name.startswith(synthesizer['name'])),
                    synthesizer['name'],
                )
            else:
                final_name = synthesizer['name']

            synthesizer['name'] = final_name
            job_tuples.append((synthesizer, dataset))

    job_args_list = []
    for synthesizer, dataset in job_tuples:
        data, metadata_dict = load_dataset(
            'single_table', dataset, limit_dataset_size=limit_dataset_size
        )
        path = paths.get(dataset.name, {}).get(synthesizer['name'], None)
        args = (
            synthesizer,
            data,
            metadata_dict,
            sdmetrics,
            detailed_results_folder,
            timeout,
            compute_quality_score,
            compute_diagnostic_score,
            compute_privacy_score,
            dataset.name,
            'single_table',
            path,
        )
        job_args_list.append(args)

    return job_args_list


def _synthesize(synthesizer_dict, real_data, metadata, synthesizer_path=None, result_writer=None):
    synthesizer = synthesizer_dict['synthesizer']
    if isinstance(synthesizer, type):
        assert issubclass(synthesizer, BaselineSynthesizer), (
            '`synthesizer` must be a synthesizer class'
        )
        synthesizer = synthesizer()
    else:
        assert issubclass(type(synthesizer), BaselineSynthesizer), (
            '`synthesizer` must be an instance of a synthesizer class.'
        )

    get_synthesizer = synthesizer.get_trained_synthesizer
    sample_from_synthesizer = synthesizer.sample_from_synthesizer
    data = real_data.copy()
    num_samples = len(data)

    tracemalloc.start()
    now = get_utc_now()
    synthesizer_obj = get_synthesizer(data, metadata)
    synthesizer_size = len(pickle.dumps(synthesizer_obj)) / N_BYTES_IN_MB
    train_now = get_utc_now()
    synthetic_data = sample_from_synthesizer(synthesizer_obj, num_samples)
    sample_now = get_utc_now()

    peak_memory = tracemalloc.get_traced_memory()[1] / N_BYTES_IN_MB
    tracemalloc.stop()
    tracemalloc.clear_traces()
    if synthesizer_path is not None and result_writer is not None:
        result_writer.write_dataframe(synthetic_data, synthesizer_path['synthetic_data'])
        result_writer.write_pickle(synthesizer_obj, synthesizer_path['synthesizer'])

    return synthetic_data, train_now - now, sample_now - train_now, synthesizer_size, peak_memory


def _compute_scores(
    metrics,
    real_data,
    synthetic_data,
    metadata,
    output,
    compute_quality_score,
    compute_diagnostic_score,
    compute_privacy_score,
    modality,
    dataset_name,
):
    metrics = metrics or []
    sdmetrics_metadata = convert_metadata_to_sdmetrics(metadata)
    if len(metrics) > 0:
        metrics, metric_kwargs = get_metrics(metrics, modality='single-table')
        scores = []
        output['scores'] = scores
        for metric_name, metric in metrics.items():
            scores.append({
                'metric': metric_name,
                'error': 'Metric Timeout',
            })
            # re-inject list to multiprocessing output
            output['scores'] = scores

            error = None
            score = None
            normalized_score = None
            start = get_utc_now()
            try:
                LOGGER.info('Computing %s on dataset %s', metric_name, dataset_name)
                metric_args = (real_data, synthetic_data, sdmetrics_metadata)
                score = metric.compute(*metric_args, **metric_kwargs.get(metric_name, {}))
                normalized_score = metric.normalize(score)
            except Exception:
                LOGGER.exception(
                    'Metric %s failed on dataset %s. Skipping.', metric_name, dataset_name
                )
                _, error = format_exception()

            scores[-1].update({
                'score': score,
                'normalized_score': normalized_score,
                'error': error,
                'metric_time': calculate_score_time(start),
            })
            # re-inject list to multiprocessing output
            output['scores'] = scores

    if compute_diagnostic_score:
        start = get_utc_now()
        if modality == 'single_table':
            diagnostic_report = SingleTableDiagnosticReport()
        else:
            diagnostic_report = MultiTableDiagnosticReport()

        diagnostic_report.generate(real_data, synthetic_data, sdmetrics_metadata, verbose=False)
        output['diagnostic_score_time'] = calculate_score_time(start)
        output['diagnostic_score'] = diagnostic_report.get_score()

    if compute_quality_score:
        start = get_utc_now()
        if modality == 'single_table':
            quality_report = SingleTableQualityReport()
        else:
            quality_report = MultiTableQualityReport()

        quality_report.generate(real_data, synthetic_data, sdmetrics_metadata, verbose=False)
        output['quality_score_time'] = calculate_score_time(start)
        output['quality_score'] = quality_report.get_score()

    if compute_privacy_score:
        start = get_utc_now()
        num_rows = len(synthetic_data)
        # parameters were determined by running experiments with all datasets
        # and tweaking num_rows_subsample and num_iterations until
        # absolute difference from real score was close (while keeping runtime low)
        # see SDGym/0397 folder in Drive
        num_rows_subsample = math.floor(num_rows * 0.60)
        num_iterations = 2
        score = DCRBaselineProtection.compute_breakdown(
            real_data=real_data,
            synthetic_data=synthetic_data,
            metadata=sdmetrics_metadata,
            num_rows_subsample=num_rows_subsample,
            num_iterations=num_iterations,
        )
        output['privacy_score_time'] = calculate_score_time(start)
        output['privacy_score'] = score.get('score')


def _score(
    synthesizer,
    data,
    metadata,
    metrics,
    output=None,
    compute_quality_score=False,
    compute_diagnostic_score=False,
    compute_privacy_score=False,
    modality=None,
    dataset_name=None,
    synthesizer_path=None,
    result_writer=None,
):
    if output is None:
        output = {}

    output['timeout'] = True  # To be deleted if there is no error
    output['error'] = 'Load Timeout'  # To be deleted if there is no error
    try:
        LOGGER.info(
            'Running %s on %s dataset %s; %s',
            synthesizer['name'],
            modality,
            dataset_name,
            used_memory(),
        )

        output['dataset_size'] = get_size_of(data) / N_BYTES_IN_MB
        # To be deleted if there is no error
        output['error'] = 'Synthesizer Timeout'
        synthetic_data, train_time, sample_time, synthesizer_size, peak_memory = _synthesize(
            synthesizer,
            data.copy(),
            metadata,
            synthesizer_path=synthesizer_path,
            result_writer=result_writer,
        )

        output['synthetic_data'] = synthetic_data
        output['train_time'] = train_time.total_seconds()
        output['sample_time'] = sample_time.total_seconds()
        output['synthesizer_size'] = synthesizer_size
        output['peak_memory'] = peak_memory

        LOGGER.info(
            'Scoring %s on %s dataset %s; %s',
            synthesizer['name'],
            modality,
            dataset_name,
            used_memory(),
        )

        # No error so far. _compute_scores tracks its own errors by metric
        del output['error']
        _compute_scores(
            metrics,
            data,
            synthetic_data,
            metadata,
            output,
            compute_quality_score,
            compute_diagnostic_score,
            compute_privacy_score,
            modality,
            dataset_name,
        )

        output['timeout'] = False  # There was no timeout

    except Exception:
        LOGGER.exception('Error running %s on dataset %s;', synthesizer['name'], dataset_name)

        exception, error = format_exception()
        output['exception'] = exception
        output['error'] = error
        output['timeout'] = False  # There was no timeout

    finally:
        LOGGER.info(
            'Finished %s on dataset %s; %s', synthesizer['name'], dataset_name, used_memory()
        )

    return output


@contextmanager
def multiprocessing_context():
    """Override multiprocessing ForkingPickler to use cloudpickle."""
    original_dump = multiprocessing.reduction.ForkingPickler.dumps
    original_load = multiprocessing.reduction.ForkingPickler.loads
    original_method = multiprocessing.get_start_method()

    multiprocessing.set_start_method('spawn', force=True)
    multiprocessing.reduction.ForkingPickler.dumps = cloudpickle.dumps
    multiprocessing.reduction.ForkingPickler.loads = cloudpickle.loads

    try:
        yield
    finally:
        # Restore original methods
        multiprocessing.set_start_method(original_method, force=True)
        multiprocessing.reduction.ForkingPickler.dumps = original_dump
        multiprocessing.reduction.ForkingPickler.loads = original_load


def _score_with_timeout(
    timeout,
    synthesizer,
    data,
    metadata,
    metrics,
    compute_quality_score=False,
    compute_diagnostic_score=False,
    compute_privacy_score=False,
    modality=None,
    dataset_name=None,
    synthesizer_path=None,
    result_writer=None,
):
    output = {} if isinstance(result_writer, S3ResultsWriter) else None
    args = (
        synthesizer,
        data,
        metadata,
        metrics,
        output,
        compute_quality_score,
        compute_diagnostic_score,
        compute_privacy_score,
        modality,
        dataset_name,
        synthesizer_path,
        result_writer,
    )
    if isinstance(result_writer, S3ResultsWriter):
        thread = threading.Thread(target=_score, args=args, daemon=True)
        thread.start()
        thread.join(timeout)
        if thread.is_alive():
            LOGGER.error('Timeout running %s on dataset %s;', synthesizer['name'], dataset_name)
            return {'timeout': True, 'error': 'Synthesizer Timeout'}

        return output

    with multiprocessing_context():
        with multiprocessing.Manager() as manager:
            output = manager.dict()
            args = args[:4] + (output,) + args[5:]  # replace output=None with manager.dict()
            process = multiprocessing.Process(target=_score, args=args)
            process.start()
            process.join(timeout)
            process.terminate()

            output = dict(output)
            if output.get('timeout'):
                LOGGER.error('Timeout running %s on dataset %s;', synthesizer['name'], dataset_name)

            return output


def _format_output(
    output,
    name,
    dataset_name,
    compute_quality_score,
    compute_diagnostic_score,
    compute_privacy_score,
    cache_dir,
):
    evaluate_time = 0
    if 'quality_score_time' in output:
        evaluate_time += output.get('quality_score_time', 0)
    if 'diagnostic_score_time' in output:
        evaluate_time += output.get('diagnostic_score_time', 0)
    if 'privacy_score_time' in output:
        evaluate_time += output.get('privacy_score_time', 0)

    for score in output.get('scores', []):
        if 'metric_time' in score and not np.isnan(score['metric_time']):
            evaluate_time += score['metric_time']

    if (
        'quality_score_time' not in output
        and 'scores' not in output
        and 'diagnostic_score_time' not in output
        and 'privacy_score_time' not in output
    ):
        evaluate_time = None

    scores = pd.DataFrame({
        'Synthesizer': [name],
        'Dataset': [dataset_name],
        'Dataset_Size_MB': [output.get('dataset_size')],
        'Train_Time': [output.get('train_time')],
        'Peak_Memory_MB': [output.get('peak_memory')],
        'Synthesizer_Size_MB': [output.get('synthesizer_size')],
        'Sample_Time': [output.get('sample_time')],
        'Evaluate_Time': [evaluate_time],
    })

    if compute_diagnostic_score:
        scores.insert(len(scores.columns), 'Diagnostic_Score', output.get('diagnostic_score'))

    if compute_quality_score:
        scores.insert(len(scores.columns), 'Quality_Score', output.get('quality_score'))

    if compute_privacy_score:
        scores.insert(len(scores.columns), 'Privacy_Score', output.get('privacy_score'))

    for score in output.get('scores', []):
        scores.insert(len(scores.columns), score['metric'], score['normalized_score'])

    if 'error' in output:
        scores['error'] = output['error']

    if cache_dir:
        cache_dir_name = str(cache_dir)
        base_path = f'{cache_dir_name}/{name}_{dataset_name}'
        if scores is not None:
            write_csv(scores, f'{base_path}_scores.csv', None, None)
        if 'synthetic_data' in output:
            synthetic_data = compress_pickle.dumps(output['synthetic_data'], compression='gzip')
            write_file(synthetic_data, f'{base_path}.data.gz', None, None)
        if 'exception' in output:
            exception = output['exception'].encode('utf-8')
            write_file(exception, f'{base_path}_error.txt', None, None)

    return scores


def _run_job(args):
    # Reset random seed
    np.random.seed()

    (
        synthesizer,
        data,
        metadata,
        metrics,
        cache_dir,
        timeout,
        compute_quality_score,
        compute_diagnostic_score,
        compute_privacy_score,
        dataset_name,
        modality,
        synthesizer_path,
        result_writer,
    ) = args

    name = synthesizer['name']
    LOGGER.info(
        'Evaluating %s on dataset %s with timeout %ss; %s',
        name,
        dataset_name,
        timeout,
        used_memory(),
    )
    output = {}
    try:
        if timeout:
            output = _score_with_timeout(
                timeout=timeout,
                synthesizer=synthesizer,
                data=data,
                metadata=metadata,
                metrics=metrics,
                compute_quality_score=compute_quality_score,
                compute_diagnostic_score=compute_diagnostic_score,
                compute_privacy_score=compute_privacy_score,
                modality=modality,
                dataset_name=dataset_name,
                synthesizer_path=synthesizer_path,
                result_writer=result_writer,
            )
        else:
            output = _score(
                synthesizer=synthesizer,
                data=data,
                metadata=metadata,
                metrics=metrics,
                compute_quality_score=compute_quality_score,
                compute_diagnostic_score=compute_diagnostic_score,
                compute_privacy_score=compute_privacy_score,
                modality=modality,
                dataset_name=dataset_name,
                synthesizer_path=synthesizer_path,
                result_writer=result_writer,
            )
    except Exception as error:
        output['exception'] = error

    scores = _format_output(
        output,
        name,
        dataset_name,
        compute_quality_score,
        compute_diagnostic_score,
        compute_privacy_score,
        cache_dir,
    )
    if synthesizer_path and result_writer:
        result_writer.write_dataframe(scores, synthesizer_path['benchmark_result'])

    return scores


def _run_on_dask(jobs, verbose):
    """Run the tasks in parallel using dask."""
    try:
        import dask
    except ImportError as ie:
        ie.msg += (
            '\n\nIt seems like `dask` is not installed.\n'
            'Please install `dask` and `distributed` using:\n'
            '\n    pip install dask distributed'
        )
        raise

    scorer = dask.delayed(_run_job)
    persisted = dask.persist(*[scorer(args) for args in jobs])
    if verbose:
        try:
            progress(persisted)
        except ValueError:
            pass

    return dask.compute(*persisted)


def _run_jobs(multi_processing_config, job_args_list, show_progress, result_writer=None):
    workers = 1
    if multi_processing_config:
        if multi_processing_config['package_name'] == 'dask':
            workers = 'dask'
            scores = _run_on_dask(job_args_list, show_progress)
        else:
            num_gpus = get_num_gpus()
            if num_gpus > 0:
                workers = num_gpus
            else:
                workers = multiprocessing.cpu_count()

    job_args_list = [job_args + (result_writer,) for job_args in job_args_list]
    if workers in (0, 1):
        scores = map(_run_job, job_args_list)
    elif workers != 'dask':
        pool = concurrent.futures.ProcessPoolExecutor(workers)
        scores = pool.map(_run_job, job_args_list)

    if show_progress:
        scores = tqdm.tqdm(scores, total=len(job_args_list), position=0, leave=True)
    else:
        scores = tqdm.tqdm(
            scores, total=len(job_args_list), file=TqdmLogger(), position=0, leave=True
        )

    if not scores:
        raise SDGymError('No valid Dataset/Synthesizer combination given.')

    scores = pd.concat(scores, ignore_index=True)
    _add_adjusted_scores(scores=scores, timeout=job_args_list[0][5])
    output_directions = job_args_list[0][-2]
    result_writer = job_args_list[0][-1]
    if output_directions and result_writer:
        path = output_directions['results']
        result_writer.write_dataframe(scores, path, append=True)

    return scores


def _get_empty_dataframe(
    compute_diagnostic_score, compute_quality_score, compute_privacy_score, sdmetrics
):
    warnings.warn('No datasets/synthesizers found.')

    scores = pd.DataFrame({
        'Synthesizer': [],
        'Dataset': [],
        'Dataset_Size_MB': [],
        'Train_Time': [],
        'Peak_Memory_MB': [],
        'Synthesizer_Size_MB': [],
        'Sample_Time': [],
        'Evaluate_Time': [],
        'Adjusted_Total_Time': [],
    })

    if compute_diagnostic_score:
        scores['Diagnostic_Score'] = []
    if compute_quality_score:
        scores['Quality_Score'] = []
        scores['Adjusted_Quality_Score'] = []
    if compute_privacy_score:
        scores['Privacy_Score'] = []
    if sdmetrics:
        for metric in sdmetrics:
            scores[metric[0]] = []

    return scores


def _directory_exists(bucket_name, s3_file_path):
    # Find the last occurrence of '/' in the file path
    last_slash_index = s3_file_path.rfind('/')
    directory_prefix = s3_file_path[: last_slash_index + 1]
    s3_client = boto3.client('s3')
    response = s3_client.list_objects_v2(Bucket=bucket_name, Prefix=directory_prefix, Delimiter='/')
    return 'Contents' in response or 'CommonPrefixes' in response


def _check_write_permissions(s3_client, bucket_name):
    s3_client = s3_client or boto3.client('s3')
    try:
        s3_client.put_object(Bucket=bucket_name, Key='__test__', Body=b'')
        write_permission = True
    except Exception:
        write_permission = False
    finally:
        # Clean up the test object
        if write_permission:
            s3_client.delete_object(Bucket=bucket_name, Key='__test__')
    return write_permission


def _create_sdgym_script(params, output_filepath):
    # Confirm the path works
    if not is_s3_path(output_filepath):
        raise ValueError("""Invalid S3 path format.
                         Expected 's3://<bucket_name>/<path_to_file>'.""")
    bucket_name, key_prefix = parse_s3_path(output_filepath)
    if not _directory_exists(bucket_name, key_prefix):
        raise ValueError(f'Directories in {key_prefix} do not exist')
    if not _check_write_permissions(None, bucket_name):
        raise ValueError('No write permissions allowed for the bucket.')

    # Add quotes to parameter strings
    if params['additional_datasets_folder']:
        params['additional_datasets_folder'] = "'" + params['additional_datasets_folder'] + "'"
    if params['detailed_results_folder']:
        params['detailed_results_folder'] = "'" + params['detailed_results_folder'] + "'"
    if params['output_filepath']:
        params['output_filepath'] = "'" + params['output_filepath'] + "'"

    # Generate the output script to run on the e2 instance
    synthesizers = params.get('synthesizers', [])
    names = []
    for synthesizer in synthesizers:
        if isinstance(synthesizer, str):
            names.append(synthesizer)
        elif hasattr(synthesizer, '__name__'):
            names.append(synthesizer.__name__)
        else:
            names.append(synthesizer.__class__.__name__)

    all_names = '", "'.join(names)
    synthesizer_string = f'synthesizers=["{all_names}"]'
    # The indentation of the string is important for the python script
    script_content = f"""import boto3
from io import StringIO
import sdgym

results = sdgym.benchmark_single_table(
    {synthesizer_string}, custom_synthesizers={params['custom_synthesizers']},
    sdv_datasets={params['sdv_datasets']}, output_filepath={params['output_filepath']},
    additional_datasets_folder={params['additional_datasets_folder']},
    limit_dataset_size={params['limit_dataset_size']},
    compute_quality_score={params['compute_quality_score']},
    compute_diagnostic_score={params['compute_diagnostic_score']},
    compute_privacy_score={params['compute_privacy_score']},
    sdmetrics={params['sdmetrics']},
    timeout={params['timeout']},
    detailed_results_folder={params['detailed_results_folder']},
    multi_processing_config={params['multi_processing_config']}
)
"""

    return script_content


def _create_instance_on_ec2(script_content):
    ec2_client = boto3.client('ec2')
    session = boto3.session.Session()
    credentials = session.get_credentials()
    print(f'This instance is being created in region: {session.region_name}')  # noqa
    escaped_script = script_content.strip().replace('"', '\\"')

    # User data script to install the library
    user_data_script = f"""#!/bin/bash
    sudo apt update -y
    sudo apt install -y python3-pip python3-venv awscli
    echo "======== Create Virtual Environment ============"
    python3 -m venv ~/env
    source ~/env/bin/activate
    echo "======== Install Dependencies in venv ============"
    pip install --upgrade pip
    pip install sdgym[all]
    pip install anyio
    echo "======== Configure AWS CLI ============"
    aws configure set aws_access_key_id {credentials.access_key}
    aws configure set aws_secret_access_key {credentials.secret_key}
    aws configure set region {session.region_name}
    echo "======== Write Script ==========="
    printf '%s\\n' "{escaped_script}" > ~/sdgym_script.py
    echo "======== Run Script ==========="
    python ~/sdgym_script.py

    echo "======== Complete ==========="
    INSTANCE_ID=$(curl -s http://169.254.169.254/latest/meta-data/instance-id)
    aws ec2 terminate-instances --instance-ids $INSTANCE_ID
    """

    response = ec2_client.run_instances(
        ImageId='ami-080e1f13689e07408',
        InstanceType='g4dn.4xlarge',
        MinCount=1,
        MaxCount=1,
        UserData=user_data_script,
        TagSpecifications=[
            {'ResourceType': 'instance', 'Tags': [{'Key': 'Name', 'Value': 'SDGym_Temp'}]}
        ],
        BlockDeviceMappings=[
            {
                'DeviceName': '/dev/sda1',
                'Ebs': {
                    'VolumeSize': 32,  # Specify the desired size in GB
                    'VolumeType': 'gp2',  # Change the volume type as needed
                },
            }
        ],
    )

    # Wait until the instance is running before terminating
    instance_id = response['Instances'][0]['InstanceId']
    waiter = ec2_client.get_waiter('instance_status_ok')
    waiter.wait(InstanceIds=[instance_id])
    print(f'Job kicked off for SDGym on {instance_id}')  # noqa


def _handle_deprecated_parameters(
    output_filepath,
    detailed_results_folder,
    multi_processing_config,
    run_on_ec2,
    output_destination,
):
    """Handle deprecated parameters and issue warnings."""
    parameters_to_deprecate = {
        'output_filepath': output_filepath,
        'detailed_results_folder': detailed_results_folder,
        'multi_processing_config': multi_processing_config,
        'run_on_ec2': run_on_ec2,
    }
    parameters = []
    old_parameters_to_save = ('output_filepath', 'detailed_results_folder')
    for name, value in parameters_to_deprecate.items():
        if value is not None and value:
            if name in old_parameters_to_save and output_destination is not None:
                raise ValueError(
                    f"The '{name}' parameter is deprecated and cannot be used together with "
                    "'output_destination'. Please use only 'output_destination' to specify "
                    'the output path.'
                )

            parameters.append(name)

    if parameters:
        parameters = "', '".join(sorted(parameters))
        message = (
            f"Parameters '{parameters}' are deprecated in the 'benchmark_single_table' "
            "function. For saving results, please use the 'output_destination' parameter."
            " For running SDGym remotely on AWS please use the 'benchmark_single_table_aws' method."
        )
        warnings.warn(message, FutureWarning)


def _validate_output_destination(output_destination, aws_keys=None):
    """Validate the output destination parameter."""
    if output_destination is None and aws_keys is None:
        return

    if aws_keys is not None:
        return _validate_aws_inputs(
            output_destination, aws_keys['aws_access_key_id'], aws_keys['aws_secret_access_key']
        )

    if not isinstance(output_destination, str):
        raise ValueError(
            'The `output_destination` parameter must be a string representing the output path.'
        )

    if is_s3_path(output_destination):
        raise ValueError(
            'The `output_destination` parameter cannot be an S3 path. '
            'Please use `benchmark_single_table_aws` instead.'
        )


def _write_metainfo_file(synthesizers, job_args_list, result_writer=None):
    jobs = [[job[-3], job[0]['name']] for job in job_args_list]
    if not job_args_list or not job_args_list[0][-1]:
        return

    output_directions = job_args_list[0][-1]
    path = output_directions['metainfo']
    stem = Path(path).stem
    match = FILE_INCREMENT_PATTERN.search(stem)
    increment = int(match.group(1)) if match else 0
    date_match = RESULTS_DATE_PATTERN.search(path)
    if not date_match:
        raise ValueError(f'Could not extract date from metainfo path: {path}')

    date_str = date_match.group(1)
    metadata = {
        'run_id': f'run_{date_str}_{increment}',
        'starting_date': datetime.today().strftime('%m_%d_%Y %H:%M:%S'),
        'completed_date': None,
        'sdgym_version': version('sdgym'),
        'jobs': jobs,
    }
    for synthesizer in synthesizers:
        if synthesizer not in SDV_SINGLE_TABLE_SYNTHESIZERS:
            ext_lib = EXTERNAL_SYNTHESIZER_TO_LIBRARY.get(synthesizer)
            if ext_lib:
                library_version = version(ext_lib)
                metadata[f'{ext_lib}_version'] = library_version
        elif 'sdv' not in metadata.keys():
            metadata['sdv_version'] = version('sdv')

    if result_writer:
        result_writer.write_yaml(metadata, path)


def _update_metainfo_file(run_file, result_writer=None):
    completed_date = datetime.today().strftime('%m_%d_%Y %H:%M:%S')
    update = {'completed_date': completed_date}
    if result_writer:
        result_writer.write_yaml(update, run_file, append=True)


def _ensure_uniform_included(synthesizers):
    if UniformSynthesizer not in synthesizers and UniformSynthesizer.__name__ not in synthesizers:
        LOGGER.info('Adding UniformSynthesizer to list of synthesizers.')
        synthesizers.append('UniformSynthesizer')


def _fill_adjusted_scores_with_none(scores):
    """Fill adjusted total time and quality score with NaN values."""
    scores['Adjusted_Total_Time'] = None
    if 'Quality_Score' in scores.columns:
        scores['Adjusted_Quality_Score'] = None

    return scores


def _add_adjusted_scores(scores, timeout):
    """Add adjusted total time and quality score based on UniformSynthesizer baseline."""
    timeout = timeout or 0
    uniform_mask = scores['Synthesizer'].str.contains('UniformSynthesizer', na=False)
    if not uniform_mask.any():
        return _fill_adjusted_scores_with_none(scores)

    uniform_row = scores.loc[uniform_mask].iloc[0]
    base_fit_time = uniform_row.get('Train_Time')
    base_sample_time = uniform_row.get('Sample_Time')
    base_quality_score = uniform_row.get('Quality_Score', None)

    if pd.isna(base_fit_time) or pd.isna(base_sample_time):
        return _fill_adjusted_scores_with_none(scores)

    fit_times = scores['Train_Time'].fillna(0)
    sample_times = scores['Sample_Time'].fillna(0)
    errors = scores.get('error', pd.Series([None] * len(scores)))

    timeout_mask = errors == 'Synthesizer Timeout'
    other_error_mask = errors.notna() & ~timeout_mask
    no_error_mask = errors.isna()
    adjusted_times = np.select(
        [timeout_mask, other_error_mask, no_error_mask],
        [
            base_fit_time + timeout + base_sample_time,
            base_fit_time + fit_times + sample_times + base_sample_time,
            base_fit_time + fit_times + sample_times,
        ],
        default=np.nan,
    )
    scores['Adjusted_Total_Time'] = adjusted_times

    if 'Quality_Score' not in scores.columns:
        return scores

    if pd.isna(base_quality_score):
        scores['Adjusted_Quality_Score'] = None
        return scores

    has_error = errors.notna()
    scores['Adjusted_Quality_Score'] = np.where(
        has_error, base_quality_score, scores['Quality_Score']
    )

    return scores


def benchmark_single_table(
    synthesizers=DEFAULT_SYNTHESIZERS,
    custom_synthesizers=None,
    sdv_datasets=DEFAULT_DATASETS,
    additional_datasets_folder=None,
    limit_dataset_size=False,
    compute_quality_score=True,
    compute_diagnostic_score=True,
    compute_privacy_score=True,
    sdmetrics=None,
    timeout=None,
    output_destination=None,
    output_filepath=None,
    detailed_results_folder=None,
    show_progress=False,
    multi_processing_config=None,
    run_on_ec2=False,
):
    """Run the SDGym benchmark on single-table datasets.

    Args:
        synthesizers (list[string]):
            The synthesizer(s) to evaluate. Defaults to ``[GaussianCopulaSynthesizer,
            CTGANSynthesizer]``. The available options are:

                - ``GaussianCopulaSynthesizer``
                - ``CTGANSynthesizer``
                - ``CopulaGANSynthesizer``
                - ``TVAESynthesizer``
                - ``RealTabFormerSynthesizer``

        custom_synthesizers (list[class] or ``None``):
            A list of custom synthesizer classes to use. These can be completely custom or
            they can be synthesizer variants (the output from ``create_single_table_synthesizer``
            or ``create_synthesizer_variant``). Defaults to ``None``.
        sdv_datasets (list[str] or ``None``):
            Names of the SDV demo datasets to use for the benchmark. Defaults to
            ``[adult, alarm, census, child, expedia_hotel_logs, insurance, intrusion, news,
            covtype]``. Use ``None`` to disable using any sdv datasets.
        additional_datasets_folder (str or ``None``):
            The path to a folder (local or an S3 bucket). Datasets found in this folder are
            run in addition to the SDV datasets. If ``None``, no additional datasets are used.
        limit_dataset_size (bool):
            Use this flag to limit the size of the datasets for faster evaluation. If ``True``,
            limit the size of every table to 1,000 rows (randomly sampled) and the first 10
            columns.
        compute_quality_score (bool):
            Whether or not to evaluate an overall quality score. Defaults to ``True``.
        compute_diagnostic_score (bool):
            Whether or not to evaluate an overall diagnostic score. Defaults to ``True``.
        compute_privacy_score (bool):
            Whether or not to evaluate an overall privacy score. Defaults to ``True``.
        sdmetrics (list[str]):
            A list of the different SDMetrics to use.
            If you'd like to input specific parameters into the metric, provide a tuple with
            the metric name followed by a dictionary of the parameters.
        timeout (int or ``None``):
            The maximum number of seconds to wait for synthetic data creation. If ``None``, no
            timeout is enforced.
        output_destination (str or ``None``):
            The path to the output directory where results will be saved. If ``None``, no
            output is saved. The results are saved with the following structure:
            output_destination/
                run_<id>.yaml
                SDGym_results_<date>/
                    results.csv
                    <dataset_name>_<date>/
                    meta.yaml
                    <synthesizer_name>/
                        synthesizer.pkl
                        synthetic_data.csv
        output_filepath (str or ``None``):
            A file path for where to write the output as a csv file. If ``None``, no output
            is written. If run_on_ec2 flag output_filepath needs to be defined and
            the filepath should be structured as: s3://{s3_bucket_name}/{path_to_file}
            Please make sure the path exists and permissions are given.
        detailed_results_folder (str or ``None``):
            The folder for where to store the intermediary results. If ``None``, do not store
            the intermediate results anywhere.
        show_progress (bool):
            Whether to use tqdm to keep track of the progress. Defaults to ``False``.
        multi_processing_config (dict or ``None``):
            The config to use if multi-processing is desired. For example,
            {
             'package_name': 'dask' or 'multiprocessing',
             'num_workers': 4
            }
        run_on_ec2 (bool):
            The flag is used to run the benchmark on an EC2 instance that will be created
            by a script using the authentication of the current user. The EC2 instance
            uses the LATEST released version of sdgym. Local changes or changes NOT
            in the released version will NOT be used in the ec2 instance.

    Returns:
        pandas.DataFrame:
            A table containing one row per synthesizer + dataset + metric.
    """
    _handle_deprecated_parameters(
        output_filepath,
        detailed_results_folder,
        multi_processing_config,
        run_on_ec2,
        output_destination,
    )
    _validate_output_destination(output_destination)
    if not synthesizers:
        synthesizers = []

    _ensure_uniform_included(synthesizers)
    result_writer = LocalResultsWriter()
    if run_on_ec2:
        print("This will create an instance for the current AWS user's account.")  # noqa
        if output_filepath is not None:
            script_content = _create_sdgym_script(dict(locals()), output_filepath)
            _create_instance_on_ec2(script_content)
        else:
            raise ValueError('In order to run on EC2, please provide an S3 folder output.')
        return None

    _validate_inputs(output_filepath, detailed_results_folder, synthesizers, custom_synthesizers)
    _create_detailed_results_directory(detailed_results_folder)
    job_args_list = _generate_job_args_list(
        limit_dataset_size,
        sdv_datasets,
        additional_datasets_folder,
        sdmetrics,
        detailed_results_folder,
        timeout,
        output_destination,
        compute_quality_score,
        compute_diagnostic_score,
        compute_privacy_score,
        synthesizers,
        custom_synthesizers,
        s3_client=None,
    )

    _write_metainfo_file(synthesizers, job_args_list, result_writer)
    if job_args_list:
        scores = _run_jobs(multi_processing_config, job_args_list, show_progress, result_writer)

    # If no synthesizers/datasets are passed, return an empty dataframe
    else:
        scores = _get_empty_dataframe(
            compute_diagnostic_score=compute_diagnostic_score,
            compute_quality_score=compute_quality_score,
            compute_privacy_score=compute_privacy_score,
            sdmetrics=sdmetrics,
        )

    if output_filepath:
        write_csv(scores, output_filepath, None, None)

    if output_destination and job_args_list:
        metainfo_filename = job_args_list[0][-1]['metainfo']
        _update_metainfo_file(metainfo_filename, result_writer)

    return scores


def _validate_aws_inputs(output_destination, aws_access_key_id, aws_secret_access_key):
    """Validate AWS S3 inputs for SDGym benchmark.

    If credentials are not provided, the S3 bucket must be public.
    """
    if not isinstance(output_destination, str):
        raise ValueError(
            'The `output_destination` parameter must be a string representing the S3 URL.'
        )

    if not output_destination.startswith('s3://'):
        raise ValueError("'output_destination' must be an S3 URL starting with 's3://'. ")

    bucket_name, _ = parse_s3_path(output_destination)
    if not bucket_name:
        raise ValueError(f'Invalid S3 URL: {output_destination}')

    config = Config(connect_timeout=30, read_timeout=300)
    if aws_access_key_id and aws_secret_access_key:
        s3_client = boto3.client(
            's3',
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=S3_REGION,
            config=config,
        )
    else:
        # No credentials provided — rely on default session
        s3_client = boto3.client('s3', config=config)

    s3_client.head_bucket(Bucket=bucket_name)
    if not _check_write_permissions(s3_client, bucket_name):
        raise PermissionError(
            f'No write permissions for the S3 bucket: {bucket_name}. '
            'Please check your AWS credentials or bucket policies.'
        )

    return s3_client


def _store_job_args_in_s3(output_destination, job_args_list, s3_client):
    """Store the job arguments in S3.

    During a run we temporarily store the job arguments in S3 to be able to
    retrieve them later in the EC2 instance. This is necessary because the
    EC2 instance does not have access to the local file system of the machine
    that initiated the run.
    The pkl file generated will be deleted after the run is completed.
    """
    parsed_url = urlparse(output_destination)
    bucket_name = parsed_url.netloc
    path = parsed_url.path.lstrip('/') if parsed_url.path else ''
    filename = os.path.basename(job_args_list[0][-1]['metainfo'])
    metainfo = os.path.splitext(filename)[0]
    job_args_key = f'job_args_list_{metainfo}.pkl'
    job_args_key = f'{path}{job_args_key}' if path else job_args_key

    serialized_data = pickle.dumps(job_args_list)
    s3_client.put_object(Bucket=bucket_name, Key=job_args_key, Body=serialized_data)

    return bucket_name, job_args_key


def _get_s3_script_content(
    access_key, secret_key, region_name, bucket_name, job_args_key, synthesizers
):
    return f"""
import boto3
import pickle
from sdgym.benchmark import _run_jobs, _write_metainfo_file, _update_metainfo_file
from io import StringIO
from sdgym.result_writer import S3ResultsWriter

s3_client = boto3.client(
    's3',
    aws_access_key_id='{access_key}',
    aws_secret_access_key='{secret_key}',
    region_name='{region_name}'
)
response = s3_client.get_object(Bucket='{bucket_name}', Key='{job_args_key}')
job_args_list = pickle.loads(response['Body'].read())
result_writer = S3ResultsWriter(s3_client=s3_client)
_write_metainfo_file({synthesizers}, job_args_list, result_writer)
scores = _run_jobs(None, job_args_list, False, result_writer=result_writer)
metainfo_filename = job_args_list[0][-1]['metainfo']
_update_metainfo_file(metainfo_filename, result_writer)
s3_client.delete_object(Bucket='{bucket_name}', Key='{job_args_key}')
"""


def _get_user_data_script(access_key, secret_key, region_name, script_content):
    return textwrap.dedent(f"""\
        #!/bin/bash
        set -e

        # Always terminate the instance when the script exits (success or failure)
        trap '
        INSTANCE_ID=$(curl -s http://169.254.169.254/latest/meta-data/instance-id);
        echo "======== Terminating EC2 instance: $INSTANCE_ID ==========";
        aws ec2 terminate-instances --instance-ids $INSTANCE_ID;
        ' EXIT

        exec > >(tee /var/log/user-data.log|logger -t user-data -s 2>/dev/console) 2>&1
        echo "======== Update and Install Dependencies ============"
        sudo apt update -y
        sudo apt install -y python3-pip python3-venv awscli
        echo "======== Configure AWS CLI ============"
        aws configure set aws_access_key_id '{access_key}'
        aws configure set aws_secret_access_key '{secret_key}'
        aws configure set default.region '{region_name}'

        echo "======== Create Virtual Environment ============"
        python3 -m venv ~/env
        source ~/env/bin/activate

        echo "======== Install Dependencies in venv ============"
        pip install --upgrade pip
        pip install sdgym[all]
        pip install s3fs

        echo "======== Write Script ==========="
        cat << 'EOF' > ~/sdgym_script.py
{script_content}
EOF

        echo "======== Run Script ==========="
        python ~/sdgym_script.py
        echo "======== Complete ==========="
        INSTANCE_ID=$(curl -s http://169.254.169.254/latest/meta-data/instance-id)
        aws ec2 terminate-instances --instance-ids $INSTANCE_ID
    """).strip()


def _run_on_aws(
    output_destination,
    synthesizers,
    s3_client,
    job_args_list,
    aws_access_key_id,
    aws_secret_access_key,
):
    bucket_name, job_args_key = _store_job_args_in_s3(output_destination, job_args_list, s3_client)
    script_content = _get_s3_script_content(
        aws_access_key_id,
        aws_secret_access_key,
        S3_REGION,
        bucket_name,
        job_args_key,
        synthesizers,
    )

    # Create a session and EC2 client using the provided S3 client's credentials
    session = boto3.session.Session(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        region_name=S3_REGION,
    )
    ec2_client = session.client('ec2')
    print(f'This instance is being created in region: {session.region_name}')  # noqa
    user_data_script = _get_user_data_script(
        aws_access_key_id, aws_secret_access_key, S3_REGION, script_content
    )
    response = ec2_client.run_instances(
        ImageId='ami-080e1f13689e07408',
        InstanceType='g4dn.4xlarge',
        MinCount=1,
        MaxCount=1,
        UserData=user_data_script,
        TagSpecifications=[
            {'ResourceType': 'instance', 'Tags': [{'Key': 'Name', 'Value': 'SDGym_Temp'}]}
        ],
        BlockDeviceMappings=[
            {
                'DeviceName': '/dev/sda1',
                'Ebs': {
                    'VolumeSize': 32,
                    'VolumeType': 'gp2',
                },
            }
        ],
    )

    # Wait until the instance is running
    instance_id = response['Instances'][0]['InstanceId']
    waiter = ec2_client.get_waiter('instance_status_ok')
    waiter.wait(InstanceIds=[instance_id])
    print(f'Job kicked off for SDGym on {instance_id}')  # noqa


def benchmark_single_table_aws(
    output_destination,
    aws_access_key_id=None,
    aws_secret_access_key=None,
    synthesizers=DEFAULT_SYNTHESIZERS,
    sdv_datasets=DEFAULT_DATASETS,
    additional_datasets_folder=None,
    limit_dataset_size=False,
    compute_quality_score=True,
    compute_diagnostic_score=True,
    compute_privacy_score=True,
    sdmetrics=None,
    timeout=None,
):
    """Run the SDGym benchmark on single-table datasets.

    Args:
        output_destination (str):
            An S3 bucket or filepath. The results output folder will be written here.
            Should be structured as:
            s3://{s3_bucket_name}/{path_to_file} or s3://{s3_bucket_name}.
        aws_access_key_id (str): The AWS access key id. Optional
        aws_secret_access_key (str): The AWS secret access key. Optional
        synthesizers (list[string]):
            The synthesizer(s) to evaluate. Defaults to
            ``[GaussianCopulaSynthesizer, CTGANSynthesizer]``. The available options
            are:
                - ``GaussianCopulaSynthesizer``
                - ``CTGANSynthesizer``
                - ``CopulaGANSynthesizer``
                - ``TVAESynthesizer``
                - ``RealTabFormerSynthesizer``
        sdv_datasets (list[str] or ``None``):
            Names of the SDV demo datasets to use for the benchmark. Defaults to
            ``[adult, alarm, census, child, expedia_hotel_logs, insurance, intrusion, news,
            covtype]``. Use ``None`` to disable using any sdv datasets.
        additional_datasets_folder (str or ``None``):
            The path to an S3 bucket. Datasets found in this folder are
            run in addition to the SDV datasets. If ``None``, no additional datasets are used.
        limit_dataset_size (bool):
            Use this flag to limit the size of the datasets for faster evaluation. If ``True``,
            limit the size of every table to 1,000 rows (randomly sampled) and the first 10
            columns.
        compute_quality_score (bool):
            Whether or not to evaluate an overall quality score. Defaults to ``True``.
        compute_diagnostic_score (bool):
            Whether or not to evaluate an overall diagnostic score. Defaults to ``True``.
        compute_privacy_score (bool):
            Whether or not to evaluate an overall privacy score. Defaults to ``True``.
        sdmetrics (list[str]):
            A list of the different SDMetrics to use.
            If you'd like to input specific parameters into the metric, provide a tuple with
            the metric name followed by a dictionary of the parameters.
        timeout (int or ``None``):
            The maximum number of seconds to wait for synthetic data creation. If ``None``, no
            timeout is enforced.

    Returns:
        pandas.DataFrame:
            A table containing one row per synthesizer + dataset + metric.
    """
    s3_client = _validate_output_destination(
        output_destination,
        aws_keys={
            'aws_access_key_id': aws_access_key_id,
            'aws_secret_access_key': aws_secret_access_key,
        },
    )
    if not synthesizers:
        synthesizers = []

    _ensure_uniform_included(synthesizers)
    job_args_list = _generate_job_args_list(
        limit_dataset_size=limit_dataset_size,
        sdv_datasets=sdv_datasets,
        additional_datasets_folder=additional_datasets_folder,
        sdmetrics=sdmetrics,
        timeout=timeout,
        output_destination=output_destination,
        compute_quality_score=compute_quality_score,
        compute_diagnostic_score=compute_diagnostic_score,
        compute_privacy_score=compute_privacy_score,
        synthesizers=synthesizers,
        detailed_results_folder=None,
        custom_synthesizers=None,
        s3_client=s3_client,
    )
    if not job_args_list:
        return _get_empty_dataframe(
            compute_diagnostic_score=compute_diagnostic_score,
            compute_quality_score=compute_quality_score,
            compute_privacy_score=compute_privacy_score,
            sdmetrics=sdmetrics,
        )

    _run_on_aws(
        output_destination=output_destination,
        synthesizers=synthesizers,
        s3_client=s3_client,
        job_args_list=job_args_list,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )



"""Results writer for SDGym benchmark."""

import io
import pickle
from abc import ABC, abstractmethod
from pathlib import Path

import pandas as pd
import plotly.graph_objects as go
import yaml
from openpyxl import load_workbook

from sdgym.s3 import parse_s3_path


class ResultsWriter(ABC):
    """Abstract base class for writing results to files."""

    @abstractmethod
    def write_dataframe(self, data, file_path, append=False):
        """Write a DataFrame to a file."""
        pass

    @abstractmethod
    def write_pickle(self, obj, file_path):
        """Write a Python object to a pickle file."""
        pass

    @abstractmethod
    def write_yaml(self, data, file_path, append=False):
        """Write data to a YAML file."""
        pass


class LocalResultsWriter:
    """Local results writer for saving results to the local filesystem."""

    def write_dataframe(self, data, file_path, append=False, index=False):
        """Write a DataFrame to a CSV file."""
        file_path = Path(file_path)
        if file_path.exists() and append:
            data.to_csv(file_path, mode='a', index=index, header=False)
        else:
            data.to_csv(file_path, mode='w', index=index, header=True)

    def process_data(self, writer, file_path, temp_images, sheet_name, obj, index=False):
        """Process a data item (DataFrame or Figure) and write it to the Excel writer."""
        if isinstance(obj, pd.DataFrame):
            obj.to_excel(writer, sheet_name=sheet_name, index=index)
        elif isinstance(obj, go.Figure):
            img_path = file_path.parent / f'{sheet_name}.png'
            obj.write_image(img_path)
            temp_images[sheet_name] = img_path

    def write_xlsx(self, data, file_path, index=False):
        """Write DataFrames to an Excel file.

        - Each DataFrame is saved as a table in its own sheet.
        - Newly written sheets are moved to the front.
        """
        file_path = Path(file_path)
        file_path.parent.mkdir(parents=True, exist_ok=True)
        if file_path.exists():
            writer = pd.ExcelWriter(
                file_path, mode='a', engine='openpyxl', if_sheet_exists='replace'
            )
        else:
            writer = pd.ExcelWriter(file_path, mode='w', engine='openpyxl')

        with writer:
            for sheet_name, df in data.items():
                df.to_excel(writer, sheet_name=sheet_name, index=index)

        wb = load_workbook(file_path)
        for sheet_name in reversed(data.keys()):
            ws = wb[sheet_name]
            wb._sheets.remove(ws)
            wb._sheets.insert(0, ws)

        wb.save(file_path)

    def write_pickle(self, obj, file_path):
        """Write a Python object to a pickle file."""
        with open(file_path, 'wb') as f:
            pickle.dump(obj, f)

    def write_yaml(self, data, file_path, append=False):
        """Write data to a YAML file."""
        file_path = Path(file_path)
        if append:
            if file_path.exists():
                with open(file_path, 'r') as f:
                    run_data = yaml.safe_load(f) or {}
                for key, value in data.items():
                    run_data[key] = value

                data = run_data

        with open(file_path, 'w') as f:
            yaml.dump(data, f)


class S3ResultsWriter(ResultsWriter):
    """Results writer for S3."""

    def __init__(self, s3_client):
        self.s3_client = s3_client

    def write_dataframe(self, data, file_path, append=False, index=False):
        """Write a DataFrame to S3 as a CSV file."""
        bucket, key = parse_s3_path(file_path)
        if append:
            try:
                response = self.s3_client.get_object(Bucket=bucket, Key=key)
                existing_data = pd.read_csv(io.BytesIO(response['Body'].read()))
                if not existing_data.empty:
                    data = pd.concat([existing_data, data], ignore_index=True)

            except Exception:
                pass  # If the file does not exist, we will create it

        csv_buffer = data.to_csv(index=index).encode()
        self.s3_client.put_object(Body=csv_buffer, Bucket=bucket, Key=key)

    def write_pickle(self, obj, file_path):
        """Write a Python object to S3 as a pickle file."""
        bucket, key = parse_s3_path(file_path)
        buffer = io.BytesIO()
        pickle.dump(obj, buffer)
        buffer.seek(0)
        self.s3_client.put_object(Body=buffer.read(), Bucket=bucket, Key=key)

    def write_yaml(self, data, file_path, append=False):
        """Write data to a YAML file in S3."""
        bucket, key = parse_s3_path(file_path)
        run_data = {}
        if append:
            try:
                response = self.s3_client.get_object(Bucket=bucket, Key=key)
                content = response['Body'].read().decode()
                run_data = yaml.safe_load(content) or {}
            except self.s3_client.exceptions.NoSuchKey:
                pass

        run_data.update(data)
        new_content = yaml.dump(run_data)
        self.s3_client.put_object(Body=new_content.encode(), Bucket=bucket, Key=key)



"""Utility functions for handling datasets."""

import json
import logging
from pathlib import Path
from zipfile import ZipFile

import numpy as np
import pandas as pd

LOGGER = logging.getLogger(__name__)


def _parse_numeric_value(value, dataset_name, field_name, target_type=float):
    """Generic parser for numeric values with logging and NaN fallback."""
    try:
        return target_type(value)
    except (ValueError, TypeError):
        LOGGER.info(
            f"Could not cast {field_name} '{value}' to {target_type.__name__} for dataset "
            f"'{dataset_name}' defaulting to NaN."
        )
        return np.nan


def _get_dataset_subset(data, metadata_dict, modality):
    """Limit the size of a dataset for faster evaluation or testing.

    This function reduces a dataset to a smaller subset by restricting the number
    of rows and columns to 1000 rows and 10 columns. It ensures that essential
    columns—such as sequence indices and keys in sequential datasets—are always retained.

    Args:
        data (pd.DataFrame):
            The dataset to be reduced.
        metadata_dict (dict):
            A dictionary containing the dataset's metadata.
        modality (str):
            The dataset modality. Must be one of: ``'single_table'``, ``'sequential'``.

    Returns:
        tuple[pd.DataFrame, dict]:
            A tuple containing:
            - The reduced dataset as a DataFrame.
            - The updated metadata dictionary reflecting any removed columns.

    Raises:
        ValueError:
            If the provided modality is ``'multi_table'``.
    """
    if modality == 'multi_table':
        raise ValueError('limit_dataset_size is not supported for multi-table datasets.')

    max_rows, max_columns = (1000, 10)
    tables = metadata_dict.get('tables', {})
    mandatory_columns = []
    table_name, table_info = next(iter(tables.items()))

    columns = table_info.get('columns', {})
    keep_columns = list(columns)
    if modality == 'sequential':
        seq_index = table_info.get('sequence_index')
        seq_key = table_info.get('sequence_key')
        mandatory_columns = [col for col in (seq_index, seq_key) if col]

    optional_columns = [col for col in columns if col not in mandatory_columns]

    # If we have too many columns, drop extras but never mandatory ones
    if len(columns) > max_columns:
        keep_count = max_columns - len(mandatory_columns)
        keep_columns = mandatory_columns + optional_columns[:keep_count]
        table_info['columns'] = {
            column_name: column_definition
            for column_name, column_definition in columns.items()
            if column_name in keep_columns
        }

    data = data[list(keep_columns)]
    data = data.sample(max_rows)
    return data, metadata_dict


def _read_zipped_data(zip_file_path, modality):
    data = {}
    with ZipFile(zip_file_path, 'r') as zf:
        for file_name in zf.namelist():
            if file_name.endswith('.csv'):
                key = Path(file_name).stem
                data[key] = _read_csv_from_zip(zf, csv_file_name=file_name)

    if modality != 'multi_table':
        data = next(iter(data.values()))

    return data


def _read_csv_from_zip(zip_file, csv_file_name):
    """Read a single CSV file from an open ZipFile and return a DataFrame."""
    with zip_file.open(csv_file_name) as csv_file:
        return pd.read_csv(csv_file, low_memory=False)


def _read_metadata_json(metadata_path):
    with open(metadata_path) as metadata_file:
        metadata_dict = json.load(metadata_file)

    return metadata_dict



"""Helpers to create SDGym synthesizer variants."""

from sdgym.synthesizers.base import BaselineSynthesizer
from sdgym.synthesizers.utils import _get_supported_synthesizers


def create_synthesizer_variant(display_name, synthesizer_class, synthesizer_parameters):
    """Create a new synthesizer variant.

    Args:
        display_name (str):
            Name of this synthesizer, used for display purposes in results.
        synthesizer_class (str):
            Name of the SDV synthesizer class to wrap.
        synthesizer_parameters (dict):
            A dictionary of the parameter names and values that will be used for the synthesizer.

    Returns:
        class:
            The synthesizer class.
    """
    if synthesizer_class not in _get_supported_synthesizers():
        raise ValueError(f"Synthesizer '{synthesizer_class}' is not a SDGym supported synthesizer.")

    base_class = BaselineSynthesizer.get_subclasses().get(synthesizer_class)
    NewSynthesizer = type(
        f'Variant:{display_name}',
        (base_class,),
        {
            '__module__': __name__,
            '_MODEL_KWARGS': synthesizer_parameters,
            '_NATIVELY_SUPPORTED': False,
        },
    )

    return NewSynthesizer


def _create_synthesizer_class(display_name, get_trained_fn, sample_fn, sample_arg_name):
    """Create a synthesizer class.

    Args:
        display_name(string):
            A string with the name of this synthesizer, used for display purposes only when
            the results are generated
        get_trained_synthesizer_fn (callable):
            A function to generate and train a synthesizer, given the real data and metadata.
        sample_from_synthesizer (callable):
            A function to sample from the given synthesizer.
        sample_arg_name (str):
            The name of the argument used to specify the number of samples to generate.
            Either 'num_samples' for single-table synthesizers, or 'scale' for multi-table
            synthesizers.

    Returns:
        class:
            The synthesizer class.
    """
    class_name = f'Custom:{display_name}'

    def get_trained_synthesizer(self, data, metadata):
        return get_trained_fn(data, metadata)

    if sample_arg_name == 'num_samples':

        def sample_from_synthesizer(self, synthesizer, num_samples):
            return sample_fn(synthesizer, num_samples)

    else:

        def sample_from_synthesizer(self, synthesizer, scale):
            return sample_fn(synthesizer, scale)

    CustomSynthesizer = type(
        class_name,
        (BaselineSynthesizer,),
        {
            '__module__': __name__,
            '_NATIVELY_SUPPORTED': False,
            'get_trained_synthesizer': get_trained_synthesizer,
            'sample_from_synthesizer': sample_from_synthesizer,
        },
    )

    globals()[class_name] = CustomSynthesizer
    return CustomSynthesizer


def create_single_table_synthesizer(
    display_name, get_trained_synthesizer_fn, sample_from_synthesizer_fn
):
    """Create a single-table synthesizer class."""
    return _create_synthesizer_class(
        display_name,
        get_trained_synthesizer_fn,
        sample_from_synthesizer_fn,
        sample_arg_name='num_samples',
    )


def create_multi_table_synthesizer(
    display_name, get_trained_synthesizer_fn, sample_from_synthesizer_fn
):
    """Create a multi-table synthesizer class."""
    return _create_synthesizer_class(
        display_name,
        get_trained_synthesizer_fn,
        sample_from_synthesizer_fn,
        sample_arg_name='scale',
    )



"""REaLTabFormer integration."""

import contextlib
import logging
from functools import partialmethod

import tqdm

from sdgym.synthesizers.base import BaselineSynthesizer


@contextlib.contextmanager
def prevent_tqdm_output():
    """Temporarily disables tqdm m."""
    tqdm.__init__ = partialmethod(tqdm.__init__, disable=True)
    try:
        yield
    finally:
        tqdm.__init__ = partialmethod(tqdm.__init__, disable=False)


class RealTabFormerSynthesizer(BaselineSynthesizer):
    """Custom wrapper for the REaLTabFormer synthesizer to make it work with SDGym."""

    LOGGER = logging.getLogger(__name__)
    _MODEL_KWARGS = None

    def _get_trained_synthesizer(self, data, metadata):
        try:
            from realtabformer import REaLTabFormer
        except Exception as exception:
            raise ValueError(
                "In order to use 'RealTabFormerSynthesizer' you have to install the extra"
                " dependencies by running  pip install sdgym['realtabformer'] "
            ) from exception

        with prevent_tqdm_output():
            model_kwargs = self._MODEL_KWARGS.copy() if self._MODEL_KWARGS else {}
            model = REaLTabFormer(model_type='tabular', **model_kwargs)
            model.fit(data)

        return model

    def _sample_from_synthesizer(self, synthesizer, n_sample):
        """Sample synthetic data with specified sample count."""
        return synthesizer.sample(n_sample)



"""Utility functions for synthesizers in SDGym."""

from sdgym.synthesizers.base import BaselineSynthesizer
from sdgym.synthesizers.sdv import _get_all_sdv_synthesizers, _get_sdv_synthesizers


def _get_sdgym_synthesizers():
    """Get SDGym synthesizers.

    Returns:
        list:
            A list of available SDGym synthesizer names.
    """
    synthesizers = BaselineSynthesizer._get_supported_synthesizers()
    sdv_synthesizer = _get_all_sdv_synthesizers()
    sdgym_synthesizer = [
        synthesizer for synthesizer in synthesizers if synthesizer not in sdv_synthesizer
    ]
    return sorted(sdgym_synthesizer)


def get_available_single_table_synthesizers():
    """List all available single-table synthesizers in SDGym.

    Returns:
        list:
            A sorted list of available single-table synthesizer names.
    """
    sdv_synthesizers = _get_sdv_synthesizers('single_table')
    sdgym_synthesizers = _get_sdgym_synthesizers()
    return sorted(sdv_synthesizers + sdgym_synthesizers)


def get_available_multi_table_synthesizers():
    """List all available multi-table synthesizers in SDGym.

    Returns:
        list:
            A sorted list of available multi-table synthesizer names.
    """
    return sorted(_get_sdv_synthesizers('multi_table'))


def _get_supported_synthesizers():
    """Get SDGym supported synthesizers.

    Returns:
        list:
            A list of available SDGym supported synthesizer names.
    """
    return BaselineSynthesizer._get_supported_synthesizers()



"""Handlers for managing SDGym benchmark results, supporting both local and S3 storage."""

import io
import operator
import os
import pickle
from abc import ABC, abstractmethod
from datetime import datetime

import pandas as pd
import yaml
from botocore.exceptions import ClientError

SYNTHESIZER_BASELINE = 'GaussianCopulaSynthesizer'
RESULTS_FOLDER_PREFIX = 'SDGym_results_'
metainfo_PREFIX = 'metainfo'
RESULTS_FILE_PREFIX = 'results'
NUM_DIGITS_DATE = 10
REGEX_SYNTHESIZER_NAME = r'\s*\(\d+\)\s*$'


class ResultsHandler(ABC):
    """Abstract base class for handling results storage and retrieval."""

    @abstractmethod
    def list(self):
        """List all runs in the results directory."""
        pass

    @abstractmethod
    def get_file_path(self, path_parts, end_filename):
        """Validate access to a specific file in the results directory."""
        pass

    @abstractmethod
    def load_synthesizer(self, file_path):
        """Load a synthesizer from a file."""
        pass

    @abstractmethod
    def load_synthetic_data(self, file_path):
        """Load synthetic data from a file."""
        pass

    @abstractmethod
    def _load_yaml_file(self, folder_name, file_name):
        """Load a YAML file from the results folder."""
        pass

    def _validate_folder_name(self, folder_name):
        """Validate that the provided folder name exists in the results directory."""
        all_folders = self.list()
        if folder_name not in all_folders:
            raise ValueError(f"Folder '{folder_name}' does not exist in the results directory.")

    def _compute_wins(self, result):
        synthesizers = result['Synthesizer'].unique()
        datasets = result['Dataset'].unique()
        result['Win'] = 0
        for dataset in datasets:
            score_baseline = result.loc[
                (result['Synthesizer'] == SYNTHESIZER_BASELINE) & (result['Dataset'] == dataset)
            ]['Quality_Score'].to_numpy()
            if score_baseline.size == 0:
                continue

            for synthesizer in synthesizers:
                loc_synthesizer = (result['Synthesizer'] == synthesizer) & (
                    result['Dataset'] == dataset
                )
                score_synthesizer = result.loc[loc_synthesizer]['Quality_Score'].to_numpy()
                result.loc[loc_synthesizer, 'Win'] = (score_synthesizer > score_baseline).astype(
                    int
                )

    def _get_summarize_table(self, folder_to_results, folder_infos):
        """Create a summary table from the results."""
        columns = []
        for folder, results in folder_to_results.items():
            date_str = folder_infos[folder]['date']
            date_obj = datetime.strptime(date_str, '%m_%d_%Y')
            column_name = (
                f'{date_str}'
                f' - # datasets: {folder_infos[folder]["# datasets"]}'
                f' - sdgym version: {folder_infos[folder]["sdgym_version"]}'
            )
            results = results.loc[results['Synthesizer'] != SYNTHESIZER_BASELINE]
            column_data = results.groupby(['Synthesizer'])['Win'].sum()
            columns.append((date_obj, column_name, column_data))

        columns.sort(key=operator.itemgetter(0))
        summarized_results = pd.DataFrame()
        for _, column_name, column_data in reversed(columns):
            summarized_results[column_name] = column_data

        summarized_results = summarized_results.fillna('-')
        summarized_results = summarized_results.reset_index()
        summarized_results = summarized_results.rename(columns={'index': 'Synthesizer'})

        return summarized_results

    def _get_column_name_infos(self, folder_to_results):
        folder_to_info = {}
        for folder, results in folder_to_results.items():
            yaml_files = self._get_results_files(folder, prefix=metainfo_PREFIX, suffix='.yaml')
            if not yaml_files:
                continue

            metainfo_info = self._load_yaml_file(folder, yaml_files[0])
            num_datasets = results.loc[
                results['Synthesizer'] == SYNTHESIZER_BASELINE, 'Dataset'
            ].nunique()
            folder_to_info[folder] = {
                'date': metainfo_info.get('starting_date')[:NUM_DIGITS_DATE],
                'sdgym_version': metainfo_info.get('sdgym_version'),
                '# datasets': num_datasets,
            }

        return folder_to_info

    def _process_results(self, results):
        """Process results to ensure they are unique and each dataset has all synthesizers."""
        aggregated_results = pd.concat(results, ignore_index=True)
        aggregated_results['Synthesizer'] = (
            aggregated_results['Synthesizer']
            .astype(str)
            .str.replace(REGEX_SYNTHESIZER_NAME, '', regex=True)
            .str.strip()
        )
        aggregated_results = aggregated_results.drop_duplicates(
            subset=['Dataset', 'Synthesizer'], keep='first'
        )
        all_synthesizers = aggregated_results['Synthesizer'].unique()
        dataset_synth_counts = aggregated_results.groupby('Dataset')['Synthesizer'].nunique()
        valid_datasets = dataset_synth_counts[dataset_synth_counts == len(all_synthesizers)].index
        filtered_results = aggregated_results[aggregated_results['Dataset'].isin(valid_datasets)]
        if filtered_results.empty:
            raise ValueError(
                'There is no dataset that has been run by all synthesizers. Cannot '
                'summarize results.'
            )

        filtered_results = filtered_results.sort_values(by=['Dataset', 'Synthesizer'])
        return filtered_results.reset_index(drop=True)

    def summarize(self, folder_name):
        """Summarize the results in the specified folder."""
        all_folders = [f for f in self.list() if f.startswith(RESULTS_FOLDER_PREFIX)]
        if folder_name not in all_folders:
            raise ValueError(f'Folder "{folder_name}" does not exist in the results directory.')

        date = pd.to_datetime(folder_name[-NUM_DIGITS_DATE:], format='%m_%d_%Y')
        folder_to_results = {}
        for folder in all_folders:
            folder_date = pd.to_datetime(folder[len(RESULTS_FOLDER_PREFIX) :], format='%m_%d_%Y')
            if folder_date > date:
                continue

            result_filenames = self._get_results_files(
                folder, prefix=RESULTS_FILE_PREFIX, suffix='.csv'
            )
            if not result_filenames:
                continue

            results = self._get_results(folder, result_filenames)
            if not results:
                continue

            aggregated_results = self._process_results(results)
            self._compute_wins(aggregated_results)
            folder_to_results[folder] = aggregated_results
            folder_infos = self._get_column_name_infos(folder_to_results)

        summarized_table = self._get_summarize_table(folder_to_results, folder_infos)

        return summarized_table, folder_to_results[folder_name]

    def load_results(self, results_folder_name):
        """Load and aggregate all the results CSV files from the specified results folder.

        Args:
            results_folder_name (str):
                The name of the results folder to load results from.

        Returns:
            pd.DataFrame:
                A DataFrame containing the results of the specified folder.
        """
        self._validate_folder_name(results_folder_name)
        result_filenames = self._get_results_files(
            results_folder_name, prefix=RESULTS_FILE_PREFIX, suffix='.csv'
        )

        return pd.concat(
            self._get_results(results_folder_name, result_filenames),
            ignore_index=True,
        )

    def load_metainfo(self, results_folder_name):
        """Load and aggregate all the metainfo YAML files from the specified results folder.

        Args:
            results_folder_name (str):
                The name of the results folder to load metainfo from.

        Returns:
            dict:
                A dictionary containing the metainfo of the specified folder.
        """
        self._validate_folder_name(results_folder_name)
        yaml_files = self._get_results_files(
            results_folder_name, prefix=metainfo_PREFIX, suffix='.yaml'
        )
        results = {}
        for yaml_file in yaml_files:
            metainfo = self._load_yaml_file(results_folder_name, yaml_file)
            run_id = metainfo.pop('run_id', None)
            results[run_id] = metainfo

        return results

    def all_runs_complete(self, folder_name):
        """Check if all runs in the specified folder are complete."""
        yaml_files = self._get_results_files(folder_name, prefix=metainfo_PREFIX, suffix='.yaml')
        if not yaml_files:
            return False

        for yaml_file in yaml_files:
            metainfo_info = self._load_yaml_file(folder_name, yaml_file)
            if metainfo_info.get('completed_date') is None:
                return False

        return True


class LocalResultsHandler(ResultsHandler):
    """Results handler for local filesystem."""

    def __init__(self, base_path):
        self.base_path = base_path

    def list(self):
        """List all runs in the local filesystem."""
        return [
            d for d in os.listdir(self.base_path) if os.path.isdir(os.path.join(self.base_path, d))
        ]

    def get_file_path(self, path_parts, end_filename):
        """Validate access to a specific file in the local filesystem."""
        full_path = os.path.join(self.base_path, *path_parts)
        if not os.path.exists(full_path):
            raise ValueError(f'Path does not exist: {full_path}')

        if not os.path.isfile(os.path.join(full_path, end_filename)):
            raise ValueError(f'File does not exist: {end_filename}')

        return os.path.join(*path_parts, end_filename)

    def load_synthesizer(self, file_path):
        """Load a synthesizer from a pickle file."""
        with open(os.path.join(self.base_path, file_path), 'rb') as f:
            return pickle.load(f)

    def load_synthetic_data(self, file_path):
        """Load synthetic data from a CSV file."""
        return pd.read_csv(os.path.join(self.base_path, file_path))

    def _get_results_files(self, folder_name, prefix, suffix):
        return [
            f
            for f in os.listdir(os.path.join(self.base_path, folder_name))
            if f.endswith(suffix) and f.startswith(prefix)
        ]

    def _get_results(self, folder_name, file_names):
        return [
            pd.read_csv(os.path.join(self.base_path, folder_name, file_name))
            for file_name in file_names
        ]

    def _load_yaml_file(self, folder_name, file_name):
        file_path = os.path.join(self.base_path, folder_name, file_name)
        with open(file_path, 'r') as f:
            return yaml.safe_load(f)


class S3ResultsHandler(ResultsHandler):
    """Results handler for AWS S3 storage."""

    def __init__(self, path, s3_client):
        self.s3_client = s3_client
        self.bucket_name = path.split('/')[2]
        self.prefix = '/'.join(path.split('/')[3:]).rstrip('/') + '/'

    def list(self):
        """List all runs in the S3 bucket."""
        response = self.s3_client.list_objects_v2(
            Bucket=self.bucket_name, Prefix=self.prefix, Delimiter='/'
        )
        return [cp['Prefix'].split('/')[-2] for cp in response.get('CommonPrefixes', [])]

    def get_file_path(self, path_parts, end_filename):
        """Validate access to a specific file in S3."""
        idx_to_structure = {0: 'Folder', 1: 'Dataset', 2: 'Synthesizer'}
        file_path = '/'.join(path_parts + [end_filename])
        previous_s3_key = self.prefix
        for idx in range(len(path_parts)):
            level_name = idx_to_structure[idx]
            current_path = '/'.join(path_parts[: idx + 1]) + '/'
            s3_key = f'{self.prefix}{current_path}'
            response = self.s3_client.list_objects_v2(
                Bucket=self.bucket_name, Prefix=s3_key, MaxKeys=1
            )

            if 'Contents' not in response:
                # If missing, fetch available items under previous level
                parent_response = self.s3_client.list_objects_v2(
                    Bucket=self.bucket_name, Prefix=previous_s3_key
                )
                available_items = set()
                if 'Contents' in parent_response:
                    for obj in parent_response['Contents']:
                        rel_path = obj['Key'][len(previous_s3_key) :]
                        if '/' in rel_path:
                            folder = rel_path.split('/')[0]
                            if folder:
                                folder = folder[: -NUM_DIGITS_DATE - 1] if idx == 1 else folder
                                available_items.add(folder)

                folder_name = path_parts[idx]
                available_list = ',\n'.join(sorted(available_items)) or 'None'
                if level_name == 'Dataset':
                    folder_name = folder_name[: -NUM_DIGITS_DATE - 1]

                if level_name == 'Folder':
                    raise ValueError(
                        f"The specified run '{folder_name}' does not exist in 'Benchmarks'. "
                        f'The available runs are:\n{available_list}'
                    )
                elif level_name == 'Dataset':
                    run_name = path_parts[0]
                    raise ValueError(
                        f"Dataset '{folder_name}' was not part of the run '{run_name}'. "
                        f'The available datasets for this run are:\n{available_list}'
                    )
                else:
                    run_name = path_parts[0]
                    dataset_name = path_parts[1][: -NUM_DIGITS_DATE - 1]
                    raise ValueError(
                        f"Synthesizer '{folder_name}' was not part of the run '{run_name}' "
                        f"for the dataset '{dataset_name}'. "
                        'The available synthesizers for this run and dataset are'
                        f':\n{available_list}'
                    )

            previous_s3_key = s3_key

        key = f'{self.prefix}{file_path}'
        try:
            self.s3_client.head_object(Bucket=self.bucket_name, Key=key)
        except ClientError as e:
            raise ValueError(
                f'File "{end_filename}" does not exist in S3 path: {self.prefix}{file_path}'
            ) from e

        return file_path

    def load_synthesizer(self, file_path):
        """Load a synthesizer from S3."""
        response = self.s3_client.get_object(
            Bucket=self.bucket_name, Key=f'{self.prefix}{file_path}'
        )
        return pickle.loads(response['Body'].read())

    def load_synthetic_data(self, file_path):
        """Load synthetic data from S3."""
        response = self.s3_client.get_object(
            Bucket=self.bucket_name, Key=f'{self.prefix}{file_path}'
        )
        return pd.read_csv(io.BytesIO(response['Body'].read()))

    def _get_results_files(self, folder_name, prefix, suffix):
        s3_prefix = f'{self.prefix}{folder_name}/'
        response = self.s3_client.list_objects_v2(Bucket=self.bucket_name, Prefix=s3_prefix)
        if 'Contents' not in response:
            return []

        return [
            obj['Key'].split('/')[-1]
            for obj in response['Contents']
            if obj['Key'].startswith(s3_prefix + prefix) and obj['Key'].endswith(suffix)
        ]

    def _get_results(self, folder_name, file_names):
        results = []
        for file_name in file_names:
            s3_key = f'{self.prefix}{folder_name}/{file_name}'
            response = self.s3_client.get_object(Bucket=self.bucket_name, Key=s3_key)
            df = pd.read_csv(io.BytesIO(response['Body'].read()))
            results.append(df)

        return results

    def _load_yaml_file(self, folder_name, file_name):
        s3_key = f'{self.prefix}{folder_name}/{file_name}'
        response = self.s3_client.get_object(Bucket=self.bucket_name, Key=s3_key)
        return yaml.safe_load(response['Body'])



"""DataIdentity module."""

import copy

from sdgym.synthesizers.base import BaselineSynthesizer


class DataIdentity(BaselineSynthesizer):
    """Trivial synthesizer.

    Returns the same exact data that is used to fit it.
    """

    def __init__(self):
        self._data = None

    def get_trained_synthesizer(self, data, metadata):
        """Get a synthesizer that has been trained on the provided data and metadata.

        Args:
            data (pandas.DataFrame):
                The data to train on.
            metadata (dict):
                The metadata dictionary.

        Returns:
            obj:
                The synthesizer object.
        """
        self._data = data
        return None

    def sample_from_synthesizer(self, synthesizer, n_samples):
        """Sample data from the provided synthesizer.

        Args:
            synthesizer (obj):
                The synthesizer object to sample data from.
            n_samples (int):
                The number of samples to create.

        Returns:
            pandas.DataFrame or dict:
                The sampled data. If single-table, should be a DataFrame. If multi-table,
                should be a dict mapping table name to DataFrame.
        """
        return copy.deepcopy(self._data)



from scripts.check_for_prereleases import get_dev_dependencies


def test_get_dev_dependencies():
    """Test get_dev_dependencies ignores regular releases."""
    # Setup
    dependencies = ['rdt>=1.1.1', 'sdv>=1.0.2']

    # Run
    dev_dependencies = get_dev_dependencies(dependency_list=dependencies)

    # Assert
    assert len(dev_dependencies) == 0


def test_get_dev_dependencies_prereleases():
    """Test get_dev_dependencies detects prereleases."""
    # Setup
    dependencies = ['rdt>=1.1.1.dev0', 'sdv>=1.0.2.rc1']

    # Run
    dev_dependencies = get_dev_dependencies(dependency_list=dependencies)

    # Assert
    assert dev_dependencies == dependencies


def test_get_dev_dependencies_url():
    """Test get_dev_dependencies detects url requirements."""
    # Setup
    dependencies = ['rdt>=1.1.1', 'sdv @ git+https://github.com/sdv-dev/sdv.git@main']

    # Run
    dev_dependencies = get_dev_dependencies(dependency_list=dependencies)

    # Assert
    assert dev_dependencies == ['sdv @ git+https://github.com/sdv-dev/sdv.git@main']



"""SDV synthesizers wrappers for SDGym."""

import logging
import sys
from importlib import import_module

from sdv import multi_table, single_table

from sdgym.synthesizers.base import BaselineSynthesizer

LOGGER = logging.getLogger(__name__)
UNSUPPORTED_SDV_SYNTHESIZERS = ['DayZSynthesizer']
MODALITY_TO_MODULE = {
    'single_table': single_table,
    'multi_table': multi_table,
}


def _validate_modality(modality):
    """Validate that the modality is correct."""
    if modality not in ['single_table', 'multi_table']:
        raise ValueError("`modality` must be one of 'single_table' or 'multi_table'.")


def _get_sdv_synthesizers(modality):
    _validate_modality(modality)
    module = MODALITY_TO_MODULE[modality]
    available_synthesizer = {name for name, cls in module.__dict__.items() if isinstance(cls, type)}
    available_synthesizer = available_synthesizer - set(UNSUPPORTED_SDV_SYNTHESIZERS)
    return sorted(available_synthesizer)


def _get_all_sdv_synthesizers():
    """Get all available SDV synthesizers."""
    synthesizers = set()
    for modality in MODALITY_TO_MODULE.keys():
        synthesizers.update(_get_sdv_synthesizers(modality))

    return sorted(synthesizers)


def _get_trained_synthesizer(self, data, metadata):
    LOGGER.info('Fitting %s', self.__class__.__name__)
    sdv_class = getattr(import_module(f'sdv.{self.modality}'), self.SDV_NAME)
    synthesizer = sdv_class(metadata=metadata, **self._MODEL_KWARGS)
    synthesizer.fit(data)
    return synthesizer


def _sample_from_synthesizer(self, synthesizer, sample_arg):
    LOGGER.info('Sampling %s', self.__class__.__name__)
    if self.modality == 'multi_table':
        return synthesizer.sample(scale=sample_arg)

    return synthesizer.sample(num_rows=sample_arg)


def _retrieve_sdv_class(sdv_name):
    current_module = sys.modules[__name__]
    if hasattr(current_module, sdv_name):
        existing_class = getattr(current_module, sdv_name)
        if isinstance(existing_class, type):
            return existing_class

    return None


def _get_modality(sdv_name):
    """Get the modality of a SDV synthesizer."""
    st_synthesizers = _get_sdv_synthesizers('single_table')
    if sdv_name in st_synthesizers:
        return 'single_table'

    mt_synthesizers = _get_sdv_synthesizers('multi_table')
    if sdv_name in mt_synthesizers:
        return 'multi_table'

    raise ValueError(f"Synthesizer '{sdv_name}' is not a SDV synthesizer.")


def _create_sdv_class(sdv_name):
    """Create a SDV synthesizer class dynamically."""
    current_module = sys.modules[__name__]
    modality = _get_modality(sdv_name)
    synthesizer_class = type(
        sdv_name,
        (BaselineSynthesizer,),
        {
            '__module__': __name__,
            'SDV_NAME': sdv_name,
            'modality': modality,
            '_MODEL_KWARGS': {},
            '_get_trained_synthesizer': _get_trained_synthesizer,
            '_sample_from_synthesizer': _sample_from_synthesizer,
        },
    )
    setattr(current_module, sdv_name, synthesizer_class)

    return synthesizer_class


def create_sdv_synthesizer_class(sdv_name):
    """Factory for dynamically creating or retrieving SDV synthesizer classes."""
    if sdv_name not in _get_all_sdv_synthesizers():
        raise ValueError(f"Synthesizer '{sdv_name}' is not a supported SDV synthesizer.")

    return _retrieve_sdv_class(sdv_name) or _create_sdv_class(sdv_name)



"""S3 module."""

import io
import logging
import pickle
from urllib.parse import urlparse

import boto3
import botocore
import pandas as pd
import yaml

S3_PREFIX = 's3://'
S3_REGION = 'us-east-1'
LOGGER = logging.getLogger(__name__)


def is_s3_path(path):
    """Determine if the given path is an s3 path.

    Args:
        path (str):
            The path, which might be an s3 path.

    Returns:
        bool:
            A boolean representing if the path is an s3 path or not.
    """
    return isinstance(path, str) and S3_PREFIX in path


def parse_s3_path(path):
    """Parse a s3 path into the bucket and key prefix.

    Args:
        path (str):
            The s3 path to parse. The expected format for the s3 path is
            `s3://<bucket-name>/path/to/dir`.

    Returns:
        tuple: (bucket_name, key_prefix)
    """
    if not path.startswith(S3_PREFIX):
        raise ValueError(f'Invalid S3 URI: {path}')

    path_without_prefix = path[len(S3_PREFIX) :]
    parts = path_without_prefix.split('/', 1)
    bucket_name = parts[0]
    key_prefix = parts[1] if len(parts) > 1 else ''

    return bucket_name, key_prefix


def get_s3_client(aws_access_key_id=None, aws_secret_access_key=None):
    """Get the boto client for interfacing with AWS s3.

    Args:
        aws_access_key_id (str):
            The access key id that will be used to communicate with
            s3, if provided.
        aws_secret_access_key (str):
            The secret access key that will be used to communicate
            with s3, if provided.

    Returns:
        boto3.session.Session.client:
            The s3 client that can be used to read / write to s3.
    """
    if aws_access_key_id is not None and aws_secret_access_key is not None:
        # credentials available
        return boto3.client(
            's3',
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=S3_REGION,
        )
    else:
        if boto3.Session().get_credentials():
            # credentials available and will be detected automatically
            config = None
        else:
            # no credentials available, make unsigned requests
            config = botocore.config.Config(signature_version=botocore.UNSIGNED)

        return boto3.client('s3', config=config)


def write_file(data_contents, path, aws_access_key_id, aws_secret_access_key):
    """Write a file to the given path with the given contents.

    If the path is an s3 directory, we will use the given aws credentials
    to write to s3.

    Args:
        data_contents (bytes):
            The contents that will be written to the file.
        path (str):
            The path to write the file to, which can be either local
            or an s3 path.
        aws_access_key_id (str):
            The access key id that will be used to communicate with s3,
            if provided.
        aws_secret_access_key (str):
            The secret access key that will be used to communicate
            with s3, if provided.

    Returns:
        none
    """
    content_encoding = ''
    write_mode = 'w'
    if path.endswith('gz') or path.endswith('gzip'):
        content_encoding = 'gzip'
        write_mode = 'wb'
    elif isinstance(data_contents, bytes):
        write_mode = 'wb'

    if is_s3_path(path):
        s3 = get_s3_client(aws_access_key_id, aws_secret_access_key)
        bucket_name, key = parse_s3_path(path)
        s3.put_object(
            Bucket=bucket_name,
            Key=key,
            Body=data_contents,
            ContentEncoding=content_encoding,
        )
    else:
        with open(path, write_mode) as f:
            if write_mode == 'w':
                f.write(data_contents.decode('utf-8'))
            else:
                f.write(data_contents)


def write_csv(data, path, aws_access_key_id, aws_secret_access_key):
    """Write a csv file to the given path with the given contents.

    If the path is an s3 directory, we will use the given aws credentials
    to write to s3.

    Args:
        data (pandas.DataFrame):
            The data that will be written to the csv file.
        path (str):
            The path to write the file to, which can be either local
            or an s3 path.
        aws_access_key_id (str):
            The access key id that will be used to communicate with s3,
            if provided.
        aws_secret_access_key (str):
            The secret access key that will be used to communicate
            with s3, if provided.

    Returns:
        none
    """
    data_contents = data.to_csv(index=False).encode('utf-8')
    write_file(data_contents, path, aws_access_key_id, aws_secret_access_key)


def _parse_s3_paths(s3_paths_dict):
    bucket = None
    keys = {}
    for k, s3_uri in s3_paths_dict.items():
        b, key = parse_s3_path(s3_uri)
        if bucket is None:
            bucket = b
        elif bucket != b:
            raise ValueError('Different buckets found in the paths dict')

        keys[k] = key

    return bucket, keys


def _upload_pickle_to_s3(obj, s3_client, bucket_name, key):
    bytes_buffer = io.BytesIO()
    pickle.dump(obj, bytes_buffer)
    bytes_buffer.seek(0)
    s3_client.put_object(Bucket=bucket_name, Key=key, Body=bytes_buffer)


def _upload_dataframe_to_s3(data, s3_client, bucket_name, key, append=False):
    """Upload a dataframe to S3, optionally appending if the file already exists."""
    if append:
        try:
            response = s3_client.get_object(Bucket=bucket_name, Key=key)
            existing_csv = response['Body'].read().decode('utf-8')
            existing_df = pd.read_csv(io.StringIO(existing_csv))
            data = pd.concat([existing_df, data], ignore_index=True)
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] != 'NoSuchKey':
                raise e
            else:
                LOGGER.info(f'File {key} does not exist, creating a new one.')

    csv_buffer = io.StringIO()
    data.to_csv(csv_buffer, index=False)
    s3_client.put_object(Bucket=bucket_name, Key=key, Body=csv_buffer.getvalue())


def _get_s3_client(output_destination, aws_access_key_id=None, aws_secret_access_key=None):
    parsed_url = urlparse(output_destination)
    bucket_name = parsed_url.netloc
    if not bucket_name:
        raise ValueError(f'Invalid S3 URL: {output_destination}')

    if aws_access_key_id and aws_secret_access_key:
        s3_client = boto3.client(
            's3',
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            region_name=S3_REGION,
        )
    else:
        s3_client = boto3.client('s3')

    s3_client.head_bucket(Bucket=bucket_name)

    return s3_client


def _read_data_from_bucket_key(s3_client, bucket_name, object_key):
    response = s3_client.get_object(Bucket=bucket_name, Key=object_key)
    return response['Body'].read()


def _list_s3_bucket_contents(s3_client, bucket_name, prefix):
    contents = []
    paginator = s3_client.get_paginator('list_objects_v2')
    for resp in paginator.paginate(Bucket=bucket_name, Prefix=prefix):
        contents.extend(resp.get('Contents', []))

    return contents


def _load_yaml_metainfo_from_s3(s3_client, bucket_name, yaml_key):
    """Load and parse YAML metainfo from an S3 key."""
    raw_data = _read_data_from_bucket_key(s3_client, bucket_name, yaml_key)
    return yaml.safe_load(raw_data) or {}



"""SDGym CLI module."""

import argparse
import gc
import json
import logging
import sys
import warnings

import humanfriendly
import pandas as pd
import tabulate
import tqdm

import sdgym
from sdgym.synthesizers.base import BaselineSynthesizer
from sdgym.utils import get_synthesizers


def _env_setup(logfile, verbosity):
    gc.enable()
    warnings.simplefilter('ignore')

    FORMAT = '%(asctime)s - %(process)d - %(levelname)s - %(name)s - %(module)s - %(message)s'
    level = (3 - verbosity) * 10
    logging.basicConfig(filename=logfile, level=level, format=FORMAT)
    logging.getLogger('sdgym').setLevel(level)
    logging.getLogger('sdmetrics').setLevel(level)
    logging.getLogger().setLevel(logging.WARN)


def _print_table(data, sort=None, reverse=False, format=None):
    if sort:
        sort_fields = sort.split(',')
        for field in sort_fields:
            data = data.sort_values(field, ascending=not reverse)

    if format:
        for field, formatter in format.items():
            data[field] = data[field].apply(formatter)

    if 'error' in data:
        error = data['error']
        if pd.isna(error).all():
            del data['error']
        else:
            long_error = error.str.len() > 30
            data.loc[long_error, 'error'] = error[long_error].str[:30] + '...'

    print(tabulate.tabulate(data, tablefmt='github', headers=data.columns, showindex=False))  # noqa: T201


def _run(args):
    _env_setup(args.logfile, args.verbose)

    if args.distributed:
        try:
            from dask.distributed import Client, LocalCluster
        except ImportError as ie:
            ie.msg += (
                '\n\nIt seems like `dask` is not installed.\n'
                'Please install `dask` and `distributed` using:\n'
                '\n    pip install dask distributed'
            )
            raise

        processes = args.workers > 1
        client = Client(
            LocalCluster(
                processes=processes,
                n_workers=args.workers,
                threads_per_worker=args.threads,
            ),
        )
        client.register_worker_callbacks(lambda: _env_setup(args.logfile, args.verbose))

    if args.jobs:
        args.jobs = json.loads(args.jobs)

    scores = sdgym.benchmark_single_table(
        synthesizers=args.synthesizers,
        sdv_datasets=args.datasets,
        sdmetrics=args.sdmetrics,
        timeout=args.timeout,
        show_progress=args.progress,
        output_filepath=args.output_path,
    )

    if args.groupby:
        scores = scores.groupby(args.groupby).mean().reset_index()

    if scores is not None:
        _print_table(scores)


def _download_datasets(args):
    _env_setup(args.logfile, args.verbose)
    datasets = args.datasets
    if not datasets:
        datasets = sdgym.datasets._get_available_datasets(
            args.bucket, args.aws_access_key_id, args.aws_secret_access_key
        )['name']

    for dataset in tqdm.tqdm(datasets):
        sdgym.datasets.load_dataset(
            dataset,
            args.datasets_path,
            args.bucket,
            args.aws_access_key_id,
            args.aws_secret_access_key,
        )


def _list_downloaded(args):
    datasets = sdgym.cli.utils.get_downloaded_datasets(args.datasets_path)
    _print_table(datasets, args.sort, args.reverse, {'size': humanfriendly.format_size})
    print(f'Found {len(datasets)} downloaded datasets')  # noqa: T201


def _list_available(args):
    datasets = sdgym.datasets._get_available_datasets(
        args.bucket, args.aws_access_key_id, args.aws_secret_access_key
    )
    _print_table(datasets, args.sort, args.reverse, {'size': humanfriendly.format_size})


def _list_synthesizers(args):
    synthesizers = BaselineSynthesizer.get_baselines()
    _print_table(pd.DataFrame(get_synthesizers(list(synthesizers))))


def _collect(args):
    sdgym.cli.collect.collect_results(
        args.input_path, args.output_file, args.aws_access_key_id, args.aws_secret_access_key
    )


def _summary(args):
    sdgym.cli.summary.make_summary_spreadsheet(
        args.input_path,
        output_path=args.output_file,
        aws_access_key_id=args.aws_access_key_id,
        aws_secret_access_key=args.aws_secret_access_key,
    )


def _get_parser():
    parser = argparse.ArgumentParser(description='SDGym Command Line Interface')
    parser.set_defaults(action=None)
    action = parser.add_subparsers(title='action')
    action.required = True

    # run
    run = action.add_parser('run', help='Run the SDGym Benchmark.')
    run.set_defaults(action=_run)

    run.add_argument(
        '-s',
        '--synthesizers',
        nargs='*',
        type=str,
        required=False,
        help='List of synthesizers to benchmark.',
    )
    run.add_argument(
        '-d',
        '--datasets',
        nargs='*',
        type=str,
        required=False,
        help='List of datasets to benchmark.',
    )
    run.add_argument(
        '-c',
        '--cache-dir',
        type=str,
        required=False,
        help='Directory where the intermediate results will be stored.',
    )
    run.add_argument(
        '-o',
        '--output-path',
        type=str,
        required=False,
        help='Path to the CSV file where the report will be dumped',
    )
    run.add_argument(
        '-m',
        '--sdmetrics',
        nargs='+',
        help='Metrics to apply. Accepts multiple names.',
    )
    run.add_argument('-b', '--bucket', help='Bucket from which to download the datasets.')
    run.add_argument('-dp--datasets-path', help='Path where datasets can be found.')
    run.add_argument(
        '-dm--modalities', nargs='+', help='Data Modalities to run. Accepts multiple names.'
    )
    run.add_argument('-i', '--iterations', type=int, default=1, help='Number of iterations.')
    run.add_argument(
        '-D', '--distributed', action='store_true', help='Distribute computation using dask.'
    )
    run.add_argument(
        '-W',
        '--workers',
        type=int,
        default=1,
        help='Number of workers to use when distributing locally.',
    )
    run.add_argument(
        '-T', '--threads', type=int, help='Number of threads to use when distributing locally.'
    )
    run.add_argument('-l', '--logfile', type=str, help='Name of the log file.')
    run.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        help='Be verbose. Repeat for increased verbosity.',
    )
    run.add_argument(
        '-p', '--progress', action='store_true', help='Print a progress bar using tqdm.'
    )
    run.add_argument(
        'run_on_ec2',
        action='store_true',
        help='Run job on created ec2 instance with environment aws variables',
    )
    run.add_argument('-t', '--timeout', type=int, help='Maximum seconds to run for each dataset.')
    run.add_argument(
        '-g', '--groupby', nargs='+', help='Group scores leaderboard by the given fields.'
    )
    run.add_argument(
        '-ak--aws-key',
        type=str,
        required=False,
        help='Aws access key ID to use when reading datasets.',
    )
    run.add_argument(
        '-as--aws-secret',
        type=str,
        required=False,
        help='Aws secret access key to use when reading datasets.',
    )
    run.add_argument(
        '-j', '--jobs', type=str, required=False, help='Serialized list of jobs to run.'
    )
    run.add_argument(
        '-mr--max-rows', type=int, help='Cap the number of rows to model from each dataset.'
    )
    run.add_argument(
        '-mc--max-columns',
        type=int,
        help='Cap the number of columns to model from each dataset.',
    )

    # download-datasets
    download = action.add_parser('download-datasets', help='Download datasets.')
    download.set_defaults(action=_download_datasets)
    download.add_argument('-b', '--bucket', help='Bucket from which to download the datasets.')
    download.add_argument(
        '-d', '--datasets', nargs='+', help='Datasets/s to be downloaded. Accepts multiple names.'
    )
    download.add_argument(
        '-dp', '--datasets-path', help='Optional path to download the datasets to.'
    )
    download.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        help='Be verbose. Repeat for increased verbosity.',
    )
    download.add_argument('-l', '--logfile', type=str, help='Name of the log file.')
    download.add_argument(
        '-ak',
        '--aws-key',
        type=str,
        required=False,
        help='Aws access key ID to use when reading datasets.',
    )
    download.add_argument(
        '-as',
        '--aws-secret',
        type=str,
        required=False,
        help='Aws secret access key to use when reading datasets.',
    )

    # list-downloaded-datasets
    list_downloaded = action.add_parser('list-downloaded', help='List downloaded datasets.')
    list_downloaded.set_defaults(action=_list_downloaded)
    list_downloaded.add_argument(
        '-s', '--sort', default='name', help='Value to sort by (name|size). Defaults to `name`.'
    )
    list_downloaded.add_argument('-r', '--reverse', action='store_true', help='Reverse the order.')
    list_downloaded.add_argument(
        '-dp', '--datasets-path', help='Path where the datasets can be found.'
    )

    # list-available-datasets
    list_available = action.add_parser(
        'list-available', help='List datasets available for download.'
    )
    list_available.set_defaults(action=_list_available)
    list_available.add_argument(
        '-s',
        '--sort',
        default='name',
        help='Value to sort by (name|size|modality). Defaults to `name`.',
    )
    list_available.add_argument('-r', '--reverse', action='store_true', help='Reverse the order.')
    list_available.add_argument(
        '-b', '--bucket', help='Bucket from which to download the datasets.'
    )
    list_available.add_argument(
        '-ak',
        '--aws-key',
        type=str,
        required=False,
        help='Aws access key ID to use when reading datasets.',
    )
    list_available.add_argument(
        '-as',
        '--aws-secret',
        type=str,
        required=False,
        help='Aws secret access key to use when reading datasets.',
    )

    # list-synthesizers
    list_available = action.add_parser(
        'list-synthesizers', help='List synthesizers available for use.'
    )
    list_available.set_defaults(action=_list_synthesizers)

    # collect
    collect = action.add_parser('collect', help='Collect sdgym results.')
    collect.set_defaults(action=_collect)
    collect.add_argument(
        '-i',
        '--input-path',
        type=str,
        required=True,
        help='Path within which to look for sdgym results.',
    )
    collect.add_argument(
        '-o', '--output-file', type=str, help='Output file containing the collected results.'
    )
    collect.add_argument(
        '-ak',
        '--aws-key',
        type=str,
        required=False,
        help='Aws access key ID to use when reading datasets.',
    )
    collect.add_argument(
        '-as',
        '--aws-secret',
        type=str,
        required=False,
        help='Aws secret access key to use when reading datasets.',
    )

    # summary
    summary = action.add_parser('summary', help='Create summary file for sdgym results.')
    summary.set_defaults(action=_summary)
    summary.add_argument(
        '-i', '--input-path', type=str, required=True, help='Path to sdgym results file.'
    )
    summary.add_argument(
        '-o',
        '--output-file',
        type=str,
        required=False,
        help='Output file containing summary xlsx doc.',
    )
    summary.add_argument(
        '-ak',
        '--aws-key',
        type=str,
        required=False,
        help='Aws access key ID to use when reading datasets.',
    )
    summary.add_argument(
        '-as',
        '--aws-secret',
        type=str,
        required=False,
        help='Aws secret access key to use when reading datasets.',
    )

    return parser


def main():
    """Run CLI."""
    pd.set_option('display.max_columns', 1000)
    pd.set_option('display.max_rows', 1000)

    parser = _get_parser()
    if len(sys.argv) < 2:
        parser.print_help()
        sys.exit(0)

    args = parser.parse_args()
    try:
        args.action(args)
    except sdgym.errors.SDGymError as error:
        print(f'ERROR: {error}')  # noqa: T201


if __name__ == '__main__':
    main()
